# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import base64
import concurrent.futures
import errno
import logging
import os
import shutil
import tempfile
import traceback
import uuid
from collections import OrderedDict
from io import IOBase
from types import TracebackType
from typing import (
    Any,
    Callable,
    Dict,
    IO,
    Iterable,
    Iterator,
    List,
    MutableMapping,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
)
from urllib.parse import urlparse

import portalocker  # type: ignore
from iopath.common.download import download
from iopath.common.event_logger import EventLogger, VTYPE
from iopath.common.non_blocking_io import NonBlockingIOManager
from typing_extensions import Protocol


__all__ = ["LazyPath", "PathManager", "get_cache_dir", "file_lock"]


def get_cache_dir(cache_dir: Optional[str] = None) -> str:
    """
    Returns a default directory to cache static files
    (usually downloaded from Internet), if None is provided.

    Args:
        cache_dir (None or str): if not None, will be returned as is.
            If None, returns the default cache directory as:

        1) $FVCORE_CACHE, if set
        2) otherwise ~/.torch/iopath_cache
    """
    if cache_dir is None:
        cache_dir = os.path.expanduser(
            os.getenv("FVCORE_CACHE", "~/.torch/iopath_cache")
        )
    try:
        g_pathmgr.mkdirs(cache_dir)
        assert os.access(cache_dir, os.W_OK)
    except (OSError, AssertionError):
        tmp_dir = os.path.join(tempfile.gettempdir(), "iopath_cache")
        logger = logging.getLogger(__name__)
        logger.warning(f"{cache_dir} is not accessible! Using {tmp_dir} instead!")
        cache_dir = tmp_dir
    return cache_dir


def file_lock(path: str):  # type: ignore
    """
    A file lock. Once entered, it is guaranteed that no one else holds the
    same lock. Others trying to enter the lock will block for 30 minutes and
    raise an exception.

    This is useful to make sure workers don't cache files to the same location.

    Args:
        path (str): a path to be locked. This function will create a lock named
            `path + ".lock"`

    Examples:

        filename = "/path/to/file"
        with file_lock(filename):
            if not os.path.isfile(filename):
                do_create_file()
    """
    dirname = os.path.dirname(path)
    try:
        os.makedirs(dirname, exist_ok=True)
    except OSError:
        # makedir is not atomic. Exceptions can happen when multiple workers try
        # to create the same dir, despite exist_ok=True.
        # When this happens, we assume the dir is created and proceed to creating
        # the lock. If failed to create the directory, the next line will raise
        # exceptions.
        pass
    return portalocker.Lock(path + ".lock", timeout=3600)  # type: ignore


# pyre-fixme[24]: Generic type `os.PathLike` expects 1 type parameter.
class LazyPath(os.PathLike):
    """
    A path that's lazily evaluated when it's used.

    Users should be careful to not use it like a str, because
    it behaves differently from a str.
    Path manipulation functions in Python such as `os.path.*` all accept
    PathLike objects already.

    It can be materialized to a str using `os.fspath`.
    """

    KEPT_ATTRIBUTES = ["__getstate__", "__setstate__"]

    def __init__(self, func: Callable[[], str]) -> None:
        """
        Args:
            func: a function that takes no arguments and returns the
                actual path as a str. It will be called at most once.
        """
        self._func = func
        self._value: Optional[str] = None

    def _get_value(self) -> str:
        if self._value is None:
            self._value = self._func()
        return self._value  # pyre-ignore

    def __fspath__(self) -> str:
        return self._get_value()

    # behave more like a str after evaluated
    def __getattr__(self, name: str):  # type: ignore
        if name in LazyPath.KEPT_ATTRIBUTES:
            # pyre-fixme[16]: `PathLike` has no attribute `__getattr__`.
            return super().__getattr__(name)
        if self._value is None:
            raise AttributeError(f"Uninitialized LazyPath has no attribute: {name}.")
        return getattr(self._value, name)

    def __getitem__(self, key):  # type: ignore
        if self._value is None:
            raise TypeError("Uninitialized LazyPath is not subscriptable.")
        return self._value[key]  # type: ignore

    def __str__(self) -> str:
        if self._value is not None:
            return self._value  # type: ignore
        else:
            return super().__str__()


class TabularIO(Protocol):
    """
    Context Manager interface to be used by PathHandler methods.
    """

    def __enter__(self) -> "TabularIO":
        ...

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        ...

    # pyre-fixme[3]: Return annotation cannot contain `Any`.
    def __iter__(self) -> Iterator[Any]:
        ...


class PathHandler(EventLogger):
    """
    PathHandler is a base class that defines common I/O functionality for a URI
    protocol. It routes I/O for a generic URI which may look like "protocol://*"
    or a canonical filepath "/foo/bar/baz".
    """

    _strict_kwargs_check = True

    def __init__(
        self,
        async_executor: Optional[concurrent.futures.Executor] = None,
    ) -> None:
        """
        When registering a `PathHandler`, the user can optionally pass in a
        `Executor` to run the asynchronous file operations.
        NOTE: For regular non-async operations of `PathManager`, there is
        no need to pass `async_executor`.

        Args:
            async_executor (optional `Executor`): Used for async file operations.
                Usage:
                ```
                    path_handler = NativePathHandler(async_executor=exe)
                    path_manager.register_handler(path_handler)
                ```
        """
        super().__init__()
        # pyre-fixme[4]: Attribute must be annotated.
        self._non_blocking_io_manager = None
        self._non_blocking_io_executor = async_executor
        try:
            from iopath.common.setup_defaults import setup_handler_defaults

            setup_handler_defaults(self)
        except ImportError:
            pass

    def _check_kwargs(self, kwargs: Dict[str, Any]) -> None:
        """
        Checks if the given arguments are empty. Throws a ValueError if strict
        kwargs checking is enabled and args are non-empty. If strict kwargs
        checking is disabled, only a warning is logged.

        Args:
            kwargs (Dict[str, Any])
        """
        if self._strict_kwargs_check:
            if len(kwargs) > 0:
                raise ValueError("Unused arguments: {}".format(kwargs))
        else:
            logger = logging.getLogger(__name__)
            for k, v in kwargs.items():
                logger.warning("[PathManager] {}={} argument ignored".format(k, v))

    def _get_supported_prefixes(self) -> List[str]:
        """
        Returns:
            List[str]: the list of URI prefixes this PathHandler can support
        """
        raise NotImplementedError()

    def _get_local_path(self, path: str, force: bool = False, **kwargs: Any) -> str:
        """
        Get a filepath which is compatible with native Python I/O such as `open`
        and `os.path`.

        If URI points to a remote resource, this function may download and cache
        the resource to local disk. In this case, the cache stays on filesystem
        (under `file_io.get_cache_dir()`) and will be used by a different run.
        Therefore this function is meant to be used with read-only resources.

        Args:
            path (str): A URI supported by this PathHandler
            force(bool): Forces a download from backend if set to True.

        Returns:
            local_path (str): a file path which exists on the local file system
        """
        raise NotImplementedError()

    def _copy_from_local(
        self, local_path: str, dst_path: str, overwrite: bool = False, **kwargs: Any
    ) -> None:
        """
        Copies a local file to the specified URI.

        If the URI is another local path, this should be functionally identical
        to copy.

        Args:
            local_path (str): a file path which exists on the local file system
            dst_path (str): A URI supported by this PathHandler
            overwrite (bool): Bool flag for forcing overwrite of existing URI

        Returns:
            status (bool): True on success
        """
        raise NotImplementedError()

    def _opent(
        self, path: str, mode: str = "r", buffering: int = 32, **kwargs: Any
    ) -> TabularIO:
        raise NotImplementedError()

    def _open(
        self, path: str, mode: str = "r", buffering: int = -1, **kwargs: Any
    ) -> Union[IO[str], IO[bytes]]:
        """
        Open a stream to a URI, similar to the built-in `open`.

        Args:
            path (str): A URI supported by this PathHandler
            mode (str): Specifies the mode in which the file is opened. It defaults
                to 'r'.
            buffering (int): An optional integer used to set the buffering policy.
                Pass 0 to switch buffering off and an integer >= 1 to indicate the
                size in bytes of a fixed-size chunk buffer. When no buffering
                argument is given, the default buffering policy depends on the
                underlying I/O implementation.

        Returns:
            file: a file-like object.
        """
        raise NotImplementedError()

    def _opena(
        self,
        path: str,
        mode: str = "r",
        callback_after_file_close: Optional[Callable[[None], None]] = None,
        buffering: int = -1,
        **kwargs: Any,
    ) -> IOBase:
        """
        Open a file with asynchronous methods. `f.write()` calls will be dispatched
        asynchronously such that the main program can continue running.
        `f.read()` is an async method that has to run in an asyncio event loop.

        NOTE: Writes to the same path are serialized so they are written in
        the same order as they were called but writes to distinct paths can
        happen concurrently.

        Usage (write, default / without callback function):
            for n in range(50):
                results = run_a_large_task(n)
                # `f` is a file-like object with asynchronous methods
                with path_manager.opena(uri, "w") as f:
                    f.write(results)            # Runs in separate thread
                # Main process returns immediately and continues to next iteration
            path_manager.async_close()

        Usage (write, advanced / with callback function):
            # To asynchronously write to storage:
            def cb():
                path_manager.copy_from_local(
                    "checkpoint.pt", uri
                )
            f = pm.opena("checkpoint.pt", "wb", callback_after_file_close=cb)
            torch.save({...}, f)
            f.close()

        Usage (read):
            async def my_function():
              return await path_manager.opena(uri, "r").read()

        Args:
            ...same args as `_open`...
            callback_after_file_close (Callable): An optional argument that can
                be passed to perform operations that depend on the asynchronous
                writes being completed. The file is first written to the local
                disk and then the callback is executed.
            buffering (int): An optional argument to set the buffer size for
                buffered asynchronous writing.

        Returns:
            file: a file-like object with asynchronous methods.
        """
        # Restrict mode until `NonBlockingIO` has async read feature.
        valid_modes = {"w", "a", "b"}
        if not all(m in valid_modes for m in mode):
            raise ValueError(f"`opena` mode must be write or append for path {path}")

        # TODO: Each `PathHandler` should set its own `self._buffered`
        # parameter and pass that in here. Until then, we assume no
        # buffering for any storage backend.
        if not self._non_blocking_io_manager:
            self._non_blocking_io_manager = NonBlockingIOManager(
                buffered=False,
                executor=self._non_blocking_io_executor,
            )

        try:
            return self._non_blocking_io_manager.get_non_blocking_io(
                path=self._get_path_with_cwd(path),
                io_obj=self._open(path, mode, **kwargs),
                callback_after_file_close=callback_after_file_close,
                buffering=buffering,
            )
        except ValueError:
            # When `_strict_kwargs_check = True`, then `open_callable`
            # will throw a `ValueError`. This generic `_opena` function
            # does not check the kwargs since it may include any `_open`
            # args like `encoding`, `ttl`, `has_user_data`, etc.
            logger = logging.getLogger(__name__)
            logger.exception(
                "An exception occurred in `NonBlockingIOManager`. This "
                "is most likely due to invalid `opena` args. Make sure "
                "they match the `open` args for the `PathHandler`."
            )
            # pyre-fixme[7]: Expected `Union[IO[bytes], IO[str]]` but got implicit
            #  return value of `None`.
            self._async_close()

    def _async_join(self, path: Optional[str] = None, **kwargs: Any) -> bool:
        """
        Ensures that desired async write threads are properly joined.

        Args:
            path (str): Pass in a file path to wait until all asynchronous
                activity for that path is complete. If no path is passed in,
                then this will wait until all asynchronous jobs are complete.

        Returns:
            status (bool): True on success
        """
        if not self._non_blocking_io_manager:
            logger = logging.getLogger(__name__)
            logger.warning(
                "This is an async feature. No threads to join because "
                "`opena` was not used."
            )
        self._check_kwargs(kwargs)
        return self._non_blocking_io_manager._join(
            self._get_path_with_cwd(path) if path else None
        )

    def _async_close(self, **kwargs: Any) -> bool:
        """
        Closes the thread pool used for the asynchronous operations.

        Returns:
            status (bool): True on success
        """
        if not self._non_blocking_io_manager:
            logger = logging.getLogger(__name__)
            logger.warning(
                "This is an async feature. No threadpool to close because "
                "`opena` was not used."
            )
        self._check_kwargs(kwargs)
        return self._non_blocking_io_manager._close_thread_pool()

    def _copy(
        self, src_path: str, dst_path: str, overwrite: bool = False, **kwargs: Any
    ) -> bool:
        """
        Copies a source path to a destination path.

        Args:
            src_path (str): A URI supported by this PathHandler
            dst_path (str): A URI supported by this PathHandler
            overwrite (bool): Bool flag for forcing overwrite of existing file

        Returns:
            status (bool): True on success
        """
        raise NotImplementedError()

    def _mv(self, src_path: str, dst_path: str, **kwargs: Any) -> bool:
        """
        Moves (renames) a source path to a destination path.

        Args:
            src_path (str): A URI supported by this PathHandler
            dst_path (str): A URI supported by this PathHandler

        Returns:
            status (bool): True on success
        """
        raise NotImplementedError()

    def _exists(self, path: str, **kwargs: Any) -> bool:
        """
        Checks if there is a resource at the given URI.

        Args:
            path (str): A URI supported by this PathHandler

        Returns:
            bool: true if the path exists
        """
        raise NotImplementedError()

    def _isfile(self, path: str, **kwargs: Any) -> bool:
        """
        Checks if the resource at the given URI is a file.

        Args:
            path (str): A URI supported by this PathHandler

        Returns:
            bool: true if the path is a file
        """
        raise NotImplementedError()

    def _isdir(self, path: str, **kwargs: Any) -> bool:
        """
        Checks if the resource at the given URI is a directory.

        Args:
            path (str): A URI supported by this PathHandler

        Returns:
            bool: true if the path is a directory
        """
        raise NotImplementedError()

    def _ls(self, path: str, **kwargs: Any) -> List[str]:
        """
        List the contents of the directory at the provided URI.

        Args:
            path (str): A URI supported by this PathHandler

        Returns:
            List[str]: list of contents in given path
        """
        raise NotImplementedError()

    def _mkdirs(self, path: str, **kwargs: Any) -> None:
        """
        Recursive directory creation function. Like mkdir(), but makes all
        intermediate-level directories needed to contain the leaf directory.
        Similar to the native `os.makedirs`.

        Args:
            path (str): A URI supported by this PathHandler
        """
        raise NotImplementedError()

    def _rm(self, path: str, **kwargs: Any) -> None:
        """
        Remove the file (not directory) at the provided URI.

        Args:
            path (str): A URI supported by this PathHandler
        """
        raise NotImplementedError()

    def _symlink(self, src_path: str, dst_path: str, **kwargs: Any) -> bool:
        """
        Symlink the src_path to the dst_path

        Args:
            src_path (str): A URI supported by this PathHandler to symlink from
            dst_path (str): A URI supported by this PathHandler to symlink to
        """
        raise NotImplementedError()

    def _set_cwd(self, path: Union[str, None], **kwargs: Any) -> bool:
        """
        Set the current working directory. PathHandler classes prepend the cwd
        to all URI paths that are handled.

        Args:
            path (str) or None: A URI supported by this PathHandler. Must be a valid
                absolute path or None to set the cwd to None.

        Returns:
            bool: true if cwd was set without errors
        """
        raise NotImplementedError()

    def _get_path_with_cwd(self, path: str) -> str:
        """
        Default implementation. PathHandler classes that provide a `_set_cwd`
        feature should also override this `_get_path_with_cwd` method.

        Args:
            path (str): A URI supported by this PathHandler.

        Returns:
            path (str): Full path with the cwd attached.
        """
        return path


class NativePathHandler(PathHandler):
    """
    Handles paths that can be accessed using Python native system calls. This
    handler uses `open()` and `os.*` calls on the given path.
    """

    # pyre-fixme[4]: Attribute must be annotated.
    _cwd = None

    def __init__(
        self,
        async_executor: Optional[concurrent.futures.Executor] = None,
    ) -> None:
        super().__init__(async_executor)

    def _get_local_path(self, path: str, force: bool = False, **kwargs: Any) -> str:
        self._check_kwargs(kwargs)
        return os.fspath(path)

    def _copy_from_local(
        self, local_path: str, dst_path: str, overwrite: bool = False, **kwargs: Any
    ) -> None:
        self._check_kwargs(kwargs)
        local_path = self._get_path_with_cwd(local_path)
        dst_path = self._get_path_with_cwd(dst_path)
        assert self._copy(
            src_path=local_path, dst_path=dst_path, overwrite=overwrite, **kwargs
        )

    def _open(
        self,
        path: str,
        mode: str = "r",
        buffering: int = -1,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
        closefd: bool = True,
        # pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
        opener: Optional[Callable] = None,
        **kwargs: Any,
    ) -> Union[IO[str], IO[bytes]]:
        """
        Open a path.

        Args:
            path (str): A URI supported by this PathHandler
            mode (str): Specifies the mode in which the file is opened. It defaults
                to 'r'.
            buffering (int): An optional integer used to set the buffering policy.
                Pass 0 to switch buffering off and an integer >= 1 to indicate the
                size in bytes of a fixed-size chunk buffer. When no buffering
                argument is given, the default buffering policy works as follows:
                    * Binary files are buffered in fixed-size chunks; the size of
                    the buffer is chosen using a heuristic trying to determine the
                    underlying device’s “block size” and falling back on
                    io.DEFAULT_BUFFER_SIZE. On many systems, the buffer will
                    typically be 4096 or 8192 bytes long.
            encoding (Optional[str]): the name of the encoding used to decode or
                encode the file. This should only be used in text mode.
            errors (Optional[str]): an optional string that specifies how encoding
                and decoding errors are to be handled. This cannot be used in binary
                mode.
            newline (Optional[str]): controls how universal newlines mode works
                (it only applies to text mode). It can be None, '', '\n', '\r',
                and '\r\n'.
            closefd (bool): If closefd is False and a file descriptor rather than
                a filename was given, the underlying file descriptor will be kept
                open when the file is closed. If a filename is given closefd must
                be True (the default) otherwise an error will be raised.
            opener (Optional[Callable]): A custom opener can be used by passing
                a callable as opener. The underlying file descriptor for the file
                object is then obtained by calling opener with (file, flags).
                opener must return an open file descriptor (passing os.open as opener
                results in functionality similar to passing None).

            See https://docs.python.org/3/library/functions.html#open for details.

        Returns:
            file: a file-like object.
        """
        self._check_kwargs(kwargs)
        return open(  # type: ignore
            self._get_path_with_cwd(path),
            mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            closefd=closefd,
            opener=opener,
        )

    def _copy(
        self, src_path: str, dst_path: str, overwrite: bool = False, **kwargs: Any
    ) -> bool:
        """
        Copies a source path to a destination path.

        Args:
            src_path (str): A URI supported by this PathHandler
            dst_path (str): A URI supported by this PathHandler
            overwrite (bool): Bool flag for forcing overwrite of existing file

        Returns:
            status (bool): True on success
        """
        self._check_kwargs(kwargs)
        src_path = self._get_path_with_cwd(src_path)
        dst_path = self._get_path_with_cwd(dst_path)
        if os.path.exists(dst_path) and not overwrite:
            logger = logging.getLogger(__name__)
            logger.error("Destination file {} already exists.".format(dst_path))
            return False

        try:
            shutil.copyfile(src_path, dst_path)
            return True
        except Exception as e:
            logger = logging.getLogger(__name__)
            logger.error("Error in file copy - {}".format(str(e)))
            return False

    def _mv(self, src_path: str, dst_path: str, **kwargs: Any) -> bool:
        """
        Moves (renames) a source path to a destination path.

        Args:
            src_path (str): A URI supported by this PathHandler
            dst_path (str): A URI supported by this PathHandler

        Returns:
            status (bool): True on success
        """
        self._check_kwargs(kwargs)
        src_path = self._get_path_with_cwd(src_path)
        dst_path = self._get_path_with_cwd(dst_path)
        if os.path.exists(dst_path):
            logger = logging.getLogger(__name__)
            logger.error("Destination file {} already exists.".format(dst_path))
            return False

        try:
            shutil.move(src_path, dst_path)
            return True
        except Exception as e:
            logger = logging.getLogger(__name__)
            logger.error("Error in move operation - {}".format(str(e)))
            return False

    def _symlink(self, src_path: str, dst_path: str, **kwargs: Any) -> bool:
        """
        Creates a symlink to the src_path at the dst_path

        Args:
            src_path (str): A URI supported by this PathHandler
            dst_path (str): A URI supported by this PathHandler

        Returns:
            status (bool): True on success
        """
        self._check_kwargs(kwargs)
        src_path = self._get_path_with_cwd(src_path)
        dst_path = self._get_path_with_cwd(dst_path)
        logger = logging.getLogger(__name__)
        if not os.path.exists(src_path):
            logger.error("Source path {} does not exist".format(src_path))
            return False
        if os.path.exists(dst_path):
            logger.error("Destination path {} already exists.".format(dst_path))
            return False
        try:
            os.symlink(src_path, dst_path)
            return True
        except Exception as e:
            logger.error("Error in symlink - {}".format(str(e)))
            return False

    def _exists(self, path: str, **kwargs: Any) -> bool:
        self._check_kwargs(kwargs)
        return os.path.exists(self._get_path_with_cwd(path))

    def _isfile(self, path: str, **kwargs: Any) -> bool:
        self._check_kwargs(kwargs)
        return os.path.isfile(self._get_path_with_cwd(path))

    def _isdir(self, path: str, **kwargs: Any) -> bool:
        self._check_kwargs(kwargs)
        return os.path.isdir(self._get_path_with_cwd(path))

    def _ls(self, path: str, **kwargs: Any) -> List[str]:
        self._check_kwargs(kwargs)
        return os.listdir(self._get_path_with_cwd(path))

    def _mkdirs(self, path: str, **kwargs: Any) -> None:
        self._check_kwargs(kwargs)
        try:
            os.makedirs(path, exist_ok=True)
        except OSError as e:
            # EEXIST it can still happen if multiple processes are creating the dir
            if e.errno != errno.EEXIST:
                raise

    def _rm(self, path: str, **kwargs: Any) -> None:
        self._check_kwargs(kwargs)
        os.remove(path)

    def _set_cwd(self, path: Union[str, None], **kwargs: Any) -> bool:
        self._check_kwargs(kwargs)
        # Remove cwd path if None
        if path is None:
            self._cwd = None
            return True

        # Make sure path is a valid Unix path
        if not os.path.exists(path):
            raise ValueError(f"{path} is not a valid Unix path")
        # Make sure path is an absolute path
        if not os.path.isabs(path):
            raise ValueError(f"{path} is not an absolute path")
        self._cwd = path
        return True

    def _get_path_with_cwd(self, path: str) -> str:
        if not path:
            return path
        return os.path.normpath(
            path if not self._cwd else os.path.join(self._cwd, path)
        )


class HTTPURLHandler(PathHandler):
    """
    Download URLs and cache them to disk.
    """

    MAX_FILENAME_LEN = 250

    def __init__(self) -> None:
        super().__init__()
        self.cache_map: Dict[str, str] = {}

    def _get_supported_prefixes(self) -> List[str]:
        return ["http://", "https://", "ftp://"]

    def _get_local_path(
        self,
        path: str,
        force: bool = False,
        cache_dir: Optional[str] = None,
        **kwargs: Any,
    ) -> str:
        """
        This implementation downloads the remote resource and caches it locally.
        The resource will only be downloaded if not previously requested.
        """
        self._check_kwargs(kwargs)
        if (
            force
            or path not in self.cache_map
            or not os.path.exists(self.cache_map[path])
        ):
            logger = logging.getLogger(__name__)
            parsed_url = urlparse(path)
            dirname = os.path.join(
                get_cache_dir(cache_dir), os.path.dirname(parsed_url.path.lstrip("/"))
            )
            filename = path.split("/")[-1]
            if len(filename) > self.MAX_FILENAME_LEN:
                filename = filename[:100] + "_" + uuid.uuid4().hex

            cached = os.path.join(dirname, filename)
            with file_lock(cached):
                if not os.path.isfile(cached):
                    logger.info("Downloading {} ...".format(path))
                    cached = download(path, dirname, filename=filename)
            logger.info("URL {} cached in {}".format(path, cached))
            self.cache_map[path] = cached
        return self.cache_map[path]

    def _open(
        self, path: str, mode: str = "r", buffering: int = -1, **kwargs: Any
    ) -> Union[IO[str], IO[bytes]]:
        """
        Open a remote HTTP path. The resource is first downloaded and cached
        locally.

        Args:
            path (str): A URI supported by this PathHandler
            mode (str): Specifies the mode in which the file is opened. It defaults
                to 'r'.
            buffering (int): Not used for this PathHandler.

        Returns:
            file: a file-like object.
        """
        self._check_kwargs(kwargs)
        assert mode in ("r", "rb"), "{} does not support open with {} mode".format(
            self.__class__.__name__, mode
        )
        assert (
            buffering == -1
        ), f"{self.__class__.__name__} does not support the `buffering` argument"
        local_path = self._get_local_path(path, force=False)
        return open(local_path, mode)


class OneDrivePathHandler(HTTPURLHandler):
    """
    Map OneDrive (short) URLs to direct download links
    """

    ONE_DRIVE_PREFIX = "https://1drv.ms/u/s!"

    def create_one_drive_direct_download(self, one_drive_url: str) -> str:
        """
        Converts a short OneDrive URI into a download link that can be used with wget

        Args:
            one_drive_url (str): A OneDrive URI supported by this PathHandler

        Returns:
            result_url (str): A direct download URI for the file
        """
        data_b64 = base64.b64encode(bytes(one_drive_url, "utf-8"))
        data_b64_string = (
            data_b64.decode("utf-8").replace("/", "_").replace("+", "-").rstrip("=")
        )
        result_url = (
            f"https://api.onedrive.com/v1.0/shares/u!{data_b64_string}/root/content"
        )
        return result_url

    def _get_supported_prefixes(self) -> List[str]:
        return [self.ONE_DRIVE_PREFIX]

    # pyre-fixme[14]: `_get_local_path` overrides method defined in `HTTPURLHandler`
    #  inconsistently.
    def _get_local_path(self, path: str, force: bool = False, **kwargs: Any) -> str:
        """
        This implementation downloads the remote resource and caches it locally.
        The resource will only be downloaded if not previously requested.
        """
        logger = logging.getLogger(__name__)
        direct_url = self.create_one_drive_direct_download(path)

        logger.info(f"URL {path} mapped to direct download link {direct_url}")

        return super()._get_local_path(os.fspath(direct_url), force=force, **kwargs)


class PathManager:
    """
    A class for users to open generic paths or translate generic paths to file names.

    path_manager.method(path) will do the following:
    1. Find a handler by checking the prefixes in `self._path_handlers`.
    2. Call handler.method(path) on the handler that's found
    """

    def __init__(self) -> None:
        self._path_handlers: MutableMapping[str, PathHandler] = OrderedDict()
        """
        Dict from path prefix to handler.
        """

        self._native_path_handler: PathHandler = NativePathHandler()
        """
        A NativePathHandler that works on posix paths. This is used as the fallback.
        """

        self._cwd: Optional[str] = None
        """
        Keeps track of the single cwd (if set).
        NOTE: Only one PathHandler can have a cwd set at a time.
        """

        self._async_handlers: Set[PathHandler] = set()
        """
        Keeps track of the PathHandler subclasses where `opena` was used so
        all of the threads can be properly joined when calling
        `PathManager.join`.
        """

        self._enable_logging = True
        """
        Flag for enabling / disabling telemetry.
        """

    # pyre-fixme[24]: Generic type `os.PathLike` expects 1 type parameter.
    def __get_path_handler(self, path: Union[str, os.PathLike]) -> PathHandler:
        """
        Finds a PathHandler that supports the given path. Falls back to the native
        PathHandler if no other handler is found.

        Args:
            path (str or os.PathLike): URI path to resource

        Returns:
            handler (PathHandler)
        """
        path = os.fspath(path)  # pyre-ignore
        for p in self._path_handlers.keys():
            if path.startswith(p):
                return self._path_handlers[p]
        return self._native_path_handler

    def __log_tmetry_keys(self, handler: PathHandler, kvs: Dict[str, VTYPE]) -> None:
        """
        Logs a dictionary of key-value pairs to a given path handler.

        Args:
            handler (PathHandler): The path handler to send the key-value pairs to.
            kvs (Dict): Dict of Key-value pairs to be logged.

        """
        if self._enable_logging:
            handler.add_keys(kvs)
            handler.add_key("handler", type(handler).__name__)
            try:
                handler.log_event()
            except Exception:
                logger = logging.getLogger(__name__)
                logger.exception(
                    "An exception occurred in telemetry logging."
                    "Disabling telemetry to prevent further exceptions."
                )
                self._enable_logging = False

    # pyre-fixme[34]: `Variable[VTYPE <: [str, int, bool, float]]` isn't present in
    #  the function's parameters.
    def __get_open_keys(self, path: str, mode: str, buffering: int) -> Dict[str, VTYPE]:
        """
        Helper function to return common set of key-value pairs applicable to open apis.

        Args:
            path (str):
            mode (str):
            buffering (int):

        """
        kvs = {}
        kvs["op"] = "open"
        if "r" in mode:
            kvs["mode"] = "read"
        elif "w" in mode:
            kvs["mode"] = "write"
        elif "a" in mode:
            kvs["mode"] = "append"
        kvs["buffering"] = buffering
        if "b" in mode:
            kvs["format"] = "binary"
        else:
            kvs["format"] = "text"
        kvs["path"] = path

        return kvs

    def opent(
        self, path: str, mode: str = "r", buffering: int = 32, **kwargs: Any
    ) -> TabularIO:
        """
        Open a tabular data source. Only reading is supported.
        The opent() returns a Python iterable collection object, compared to bytes/text data with open()

        Args:
            path (str): A URI supported by this PathHandler
            mode (str): Specifies the mode in which the file is opened. It defaults
                to 'r'
            buffering (int): number of rows fetched and cached

        Returns:
            A TabularIO context manager object
        """
        return self.__get_path_handler(path)._opent(path, mode, buffering, **kwargs)

    def open(
        self, path: str, mode: str = "r", buffering: int = -1, **kwargs: Any
    ) -> Union[IO[str], IO[bytes]]:
        """
        Open a stream to a URI, similar to the built-in `open`.

        Args:
            path (str): A URI supported by this PathHandler
            mode (str): Specifies the mode in which the file is opened. It defaults
                to 'r'.
            buffering (int): An optional integer used to set the buffering policy.
                Pass 0 to switch buffering off and an integer >= 1 to indicate the
                size in bytes of a fixed-size chunk buffer. When no buffering
                argument is given, the default buffering policy depends on the
                underlying I/O implementation.

        Returns:
            file: a file-like object.
        """
        handler = self.__get_path_handler(path)
        # pass enable mode to handler that will be logging
        # read, write operations separately.
        handler.set_logging(self._enable_logging)
        bret = handler._open(path, mode, buffering=buffering, **kwargs)  # type: ignore

        kvs = self.__get_open_keys(path, mode, buffering)
        self.__log_tmetry_keys(handler, kvs)
        return bret

    def opena(
        self,
        path: str,
        mode: str = "r",
        buffering: int = -1,
        callback_after_file_close: Optional[Callable[[None], None]] = None,
        **kwargs: Any,
    ) -> IOBase:
        """
        Open a file with asynchronous methods. `f.write()` calls will be dispatched
        asynchronously such that the main program can continue running.
        `f.read()` is an async method that has to run in an asyncio event loop.

        NOTE: Writes to the same path are serialized so they are written in
        the same order as they were called but writes to distinct paths can
        happen concurrently.

        Usage (write, default / without callback function):
            for n in range(50):
                results = run_a_large_task(n)
                # `f` is a file-like object with asynchronous methods
                with path_manager.opena(uri, "w") as f:
                    f.write(results)            # Runs in separate thread
                # Main process returns immediately and continues to next iteration
            path_manager.async_close()

        Usage (write, advanced / with callback function):
            # To asynchronously write to storage:
            def cb():
                path_manager.copy_from_local("checkpoint.pt", uri)
            f = pm.opena("checkpoint.pt", "wb", callback_after_file_close=cb)
            torch.save({...}, f)
            f.close()

        Usage (read):
            async def my_function():
              return await path_manager.opena(uri, "r").read()

        Args:
            ...
            callback_after_file_close (Callable): Only used in write mode. An
                optional argument that can be passed to perform operations that
                depend on the asynchronous writes being completed. The file is
                first written to the local disk and then the callback is
                executed.

        Returns:
            file: a file-like object with asynchronous methods.
        """
        if "w" in mode:
            kwargs["callback_after_file_close"] = callback_after_file_close
            kwargs["buffering"] = buffering
        non_blocking_io = self.__get_path_handler(path)._opena(
            path,
            mode,
            **kwargs,
        )
        if "w" in mode:
            # Keep track of the path handlers where `opena` is used so that all of the
            # threads can be properly joined on `PathManager.join`.
            self._async_handlers.add(self.__get_path_handler(path))
        return non_blocking_io

    def async_join(self, *paths: str, **kwargs: Any) -> bool:
        """
        Ensures that desired async write threads are properly joined.

        Usage:
            Wait for asynchronous methods operating on specific file paths to
            complete.
                async_join("path/to/file1.txt")
                async_join("path/to/file2.txt", "path/to/file3.txt")
            Wait for all asynchronous methods to complete.
                async_join()

        Args:
            *paths (str): Pass in any number of file paths and `async_join` will wait
                until all asynchronous activity for those paths is complete. If no
                paths are passed in, then `async_join` will wait until all asynchronous
                jobs are complete.

        Returns:
            status (bool): True on success
        """
        success = True
        if not paths:  # Join all.
            for handler in self._async_handlers:
                success = handler._async_join(**kwargs) and success
        else:  # Join specific paths.
            for path in paths:
                success = (
                    self.__get_path_handler(path)._async_join(path, **kwargs)
                    and success
                )
        return success

    def async_close(self, **kwargs: Any) -> bool:
        """
        `async_close()` must be called at the very end of any script that uses the
        asynchronous `opena` feature. This calls `async_join()` first and then closes
        the thread pool used for the asynchronous operations.

        Returns:
            status (bool): True on success
        """
        success = self.async_join(**kwargs)
        for handler in self._async_handlers:
            success = handler._async_close(**kwargs) and success
        self._async_handlers.clear()
        return success

    def copy(
        self, src_path: str, dst_path: str, overwrite: bool = False, **kwargs: Any
    ) -> bool:
        """
        Copies a source path to a destination path.

        Args:
            src_path (str): A URI supported by this PathHandler
            dst_path (str): A URI supported by this PathHandler
            overwrite (bool): Bool flag for forcing overwrite of existing file

        Returns:
            status (bool): True on success
        """

        if self.__get_path_handler(src_path) != self.__get_path_handler(  # type: ignore
            dst_path
        ):
            return self._copy_across_handlers(src_path, dst_path, overwrite, **kwargs)

        handler = self.__get_path_handler(src_path)
        bret = handler._copy(src_path, dst_path, overwrite, **kwargs)
        kvs = {"op": "copy", "path": src_path, "dst_path": dst_path}
        self.__log_tmetry_keys(handler, kvs)
        return bret

    def mv(self, src_path: str, dst_path: str, **kwargs: Any) -> bool:
        """
        Moves (renames) a source path supported by NativePathHandler to
        a destination path.

        Args:
            src_path (str): A URI supported by NativePathHandler
            dst_path (str): A URI supported by NativePathHandler

        Returns:
            status (bool): True on success
        Exception:
            Asserts if both the src and dest paths are not supported by
            NativePathHandler.
        """

        # Moving across handlers is not supported.
        assert self.__get_path_handler(  # type: ignore
            src_path
        ) == self.__get_path_handler(
            dst_path
        ), "Src and dest paths must be supported by the same path handler."
        handler = self.__get_path_handler(src_path)
        bret = handler._mv(src_path, dst_path, **kwargs)
        kvs = {"op": "mv", "path": src_path, "dst_path": dst_path}
        self.__log_tmetry_keys(handler, kvs)
        return bret

    def get_local_path(self, path: str, force: bool = False, **kwargs: Any) -> str:
        """
        Get a filepath which is compatible with native Python I/O such as `open`
        and `os.path`.

        If URI points to a remote resource, this function may download and cache
        the resource to local disk.

        Args:
            path (str): A URI supported by this PathHandler
            force(bool): Forces a download from backend if set to True.

        Returns:
            local_path (str): a file path which exists on the local file system
        """
        path = os.fspath(path)
        handler = self.__get_path_handler(path)  # type: ignore
        try:
            bret = handler._get_local_path(path, force=force, **kwargs)
        except TypeError:
            bret = handler._get_local_path(path, **kwargs)
        kvs = {"op": "get_local_path", "path": path, "force": force}
        # pyre-fixme[6]: For 2nd param expected `Dict[str, Variable[VTYPE <: [str,
        #  int, bool, float]]]` but got `Dict[str, Union[bool, str]]`.
        self.__log_tmetry_keys(handler, kvs)
        return bret

    def copy_from_local(
        self, local_path: str, dst_path: str, overwrite: bool = False, **kwargs: Any
    ) -> bool:
        """
        Copies a local file to the specified URI.

        If the URI is another local path, this should be functionally identical
        to copy.

        Args:
            local_path (str): a file path which exists on the local file system
            dst_path (str): A URI supported by this PathHandler
            overwrite (bool): Bool flag for forcing overwrite of existing URI

        Returns:
            status (bool): True on success
        """
        assert os.path.exists(local_path), f"local_path = {local_path}"
        handler = self.__get_path_handler(dst_path)

        kvs = {
            "op": "copy_from_local",
            "path": local_path,
            "dst_path": dst_path,
            "overwrite": overwrite,
        }
        bret = handler._copy_from_local(
            local_path=local_path, dst_path=dst_path, overwrite=overwrite, **kwargs
        )
        # pyre-fixme[6]: For 2nd param expected `Dict[str, Variable[VTYPE <: [str,
        #  int, bool, float]]]` but got `Dict[str, Union[bool, str]]`.
        self.__log_tmetry_keys(handler, kvs)
        # pyre-fixme[7]: Expected `bool` but got `None`.
        return bret

    def exists(self, path: str, **kwargs: Any) -> bool:
        """
        Checks if there is a resource at the given URI.

        Args:
            path (str): A URI supported by this PathHandler

        Returns:
            bool: true if the path exists
        """
        handler = self.__get_path_handler(path)
        bret = handler._exists(path, **kwargs)  # type: ignore
        kvs = {"op": "exists", "path": path}
        self.__log_tmetry_keys(handler, kvs)
        return bret

    def isfile(self, path: str, **kwargs: Any) -> bool:
        """
        Checks if there the resource at the given URI is a file.

        Args:
            path (str): A URI supported by this PathHandler

        Returns:
            bool: true if the path is a file
        """
        handler = self.__get_path_handler(path)
        bret = handler._isfile(path, **kwargs)  # type: ignore
        kvs = {"op": "isfile", "path": path}
        self.__log_tmetry_keys(handler, kvs)
        return bret

    def isdir(self, path: str, **kwargs: Any) -> bool:
        """
        Checks if the resource at the given URI is a directory.

        Args:
            path (str): A URI supported by this PathHandler

        Returns:
            bool: true if the path is a directory
        """
        handler = self.__get_path_handler(path)
        bret = handler._isdir(path, **kwargs)  # type: ignore
        kvs = {"op": "isdir", "path": path}
        self.__log_tmetry_keys(handler, kvs)
        return bret

    def ls(self, path: str, **kwargs: Any) -> List[str]:
        """
        List the contents of the directory at the provided URI.

        Args:
            path (str): A URI supported by this PathHandler

        Returns:
            List[str]: list of contents in given path
        """
        return self.__get_path_handler(path)._ls(path, **kwargs)

    def mkdirs(self, path: str, **kwargs: Any) -> None:
        """
        Recursive directory creation function. Like mkdir(), but makes all
        intermediate-level directories needed to contain the leaf directory.
        Similar to the native `os.makedirs`.

        Args:
            path (str): A URI supported by this PathHandler
        """
        handler = self.__get_path_handler(path)
        bret = handler._mkdirs(path, **kwargs)  # type: ignore
        kvs = {"op": "mkdirs", "path": path}
        self.__log_tmetry_keys(handler, kvs)
        return bret

    def rm(self, path: str, **kwargs: Any) -> None:
        """
        Remove the file (not directory) at the provided URI.

        Args:
            path (str): A URI supported by this PathHandler
        """
        handler = self.__get_path_handler(path)
        bret = handler._rm(path, **kwargs)  # type: ignore
        kvs = {"op": "rm", "path": path}
        self.__log_tmetry_keys(handler, kvs)
        return bret

    def symlink(self, src_path: str, dst_path: str, **kwargs: Any) -> bool:
        """Symlink the src_path to the dst_path

        Args:
            src_path (str): A URI supported by this PathHandler to symlink from
            dst_path (str): A URI supported by this PathHandler to symlink to
        """
        # Copying across handlers is not supported.
        assert self.__get_path_handler(  # type: ignore
            src_path
        ) == self.__get_path_handler(dst_path)
        handler = self.__get_path_handler(src_path)
        bret = handler._symlink(src_path, dst_path, **kwargs)  # type: ignore
        kvs = {"op": "symlink", "path": src_path, "dst_path": dst_path}
        self.__log_tmetry_keys(handler, kvs)
        return bret

    def set_cwd(self, path: Union[str, None], **kwargs: Any) -> bool:
        """
        Set the current working directory. PathHandler classes prepend the cwd
        to all URI paths that are handled.

        Args:
            path (str) or None: A URI supported by this PathHandler. Must be a valid
                absolute Unix path or None to set the cwd to None.

        Returns:
            bool: true if cwd was set without errors
        """
        if path is None and self._cwd is None:
            return True
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[typing.Any], str]`
        #  but got `Optional[str]`.
        handler = self.__get_path_handler(path or self._cwd)
        if self.__get_path_handler(path or self._cwd)._set_cwd(path, **kwargs):  # type: ignore
            self._cwd = path
            bret = True
        else:
            bret = False
        kvs = {"op": "set_cwd", "path": path}
        # pyre-fixme[6]: For 2nd param expected `Dict[str, Variable[VTYPE <: [str,
        #  int, bool, float]]]` but got `Dict[str, Optional[str]]`.
        self.__log_tmetry_keys(handler, kvs)
        return bret

    def register_handler(
        self, handler: PathHandler, allow_override: bool = True
    ) -> None:
        """
        Register a path handler associated with `handler._get_supported_prefixes`
        URI prefixes.

        Args:
            handler (PathHandler)
            allow_override (bool): allow overriding existing handler for prefix
        """
        logger = logging.getLogger(__name__)
        assert isinstance(handler, PathHandler), handler

        # Allow override of `NativePathHandler` which is automatically
        # instantiated by `PathManager`.
        if isinstance(handler, NativePathHandler):
            if allow_override:
                self._native_path_handler = handler
            else:
                raise ValueError(
                    "`NativePathHandler` is registered by default. Use the "
                    "`allow_override=True` kwarg to override it."
                )
            return

        for prefix in handler._get_supported_prefixes():
            if prefix not in self._path_handlers:
                self._path_handlers[prefix] = handler
                continue

            old_handler_type = type(self._path_handlers[prefix])
            if allow_override:
                # if using the global PathManager, show the warnings
                global g_pathmgr
                if self == g_pathmgr:
                    logger.warning(
                        f"[PathManager] Attempting to register prefix '{prefix}' from "
                        "the following call stack:\n"
                        + "".join(traceback.format_stack(limit=5))
                        # show the most recent callstack
                    )
                    logger.warning(
                        f"[PathManager] Prefix '{prefix}' is already registered "
                        f"by {old_handler_type}. We will override the old handler. "
                        "To avoid such conflicts, create a project-specific PathManager "
                        "instead."
                    )
                self._path_handlers[prefix] = handler
            else:
                raise KeyError(
                    f"[PathManager] Prefix '{prefix}' already registered by {old_handler_type}!"
                )

        # Sort path handlers in reverse order so longer prefixes take priority,
        # eg: http://foo/bar before http://foo
        self._path_handlers = OrderedDict(
            sorted(self._path_handlers.items(), key=lambda t: t[0], reverse=True)
        )

    def set_strict_kwargs_checking(self, enable: bool) -> None:
        """
        Toggles strict kwargs checking. If enabled, a ValueError is thrown if any
        unused parameters are passed to a PathHandler function. If disabled, only
        a warning is given.

        With a centralized file API, there's a tradeoff of convenience and
        correctness delegating arguments to the proper I/O layers. An underlying
        `PathHandler` may support custom arguments which should not be statically
        exposed on the `PathManager` function. For example, a custom `HTTPURLHandler`
        may want to expose a `cache_timeout` argument for `open()` which specifies
        how old a locally cached resource can be before it's refetched from the
        remote server. This argument would not make sense for a `NativePathHandler`.
        If strict kwargs checking is disabled, `cache_timeout` can be passed to
        `PathManager.open` which will forward the arguments to the underlying
        handler. By default, checking is enabled since it is innately unsafe:
        multiple `PathHandler`s could reuse arguments with different semantic
        meanings or types.

        Args:
            enable (bool)
        """
        self._native_path_handler._strict_kwargs_check = enable
        for handler in self._path_handlers.values():
            handler._strict_kwargs_check = enable

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def set_logging(self, enable_logging=True):
        self._enable_logging = enable_logging

    def _copy_across_handlers(
        self, src_path: str, dst_path: str, overwrite: bool, **kwargs: Any
    ) -> bool:
        src_handler = self.__get_path_handler(src_path)
        assert src_handler._get_local_path is not None
        dst_handler = self.__get_path_handler(dst_path)
        assert dst_handler._copy_from_local is not None

        local_file = src_handler._get_local_path(src_path, **kwargs)
        # pyre-fixme[7]: Expected `bool` but got `None`.
        return dst_handler._copy_from_local(
            local_file, dst_path, overwrite=overwrite, **kwargs
        )


"""

History: at first, PathManager is part of detectron2 *only*, and therefore
does not consider cross-projects conflict issues. It was designed to provide
a global interface with static methods.

It is later used by more projects and moved to iopath to faciliate more use
across projects and lead to some conflicts.

Now 'PathManager' class has been changed to provide a per-project interface.
This means that any project using PathManager will first have to instantiate the
class and call methods on the instantiated object.

  example:

  _pathmgr = PathManager()
  _pathmgr.register_handler(<a PathHandler instance>)
  _pathmgr.mkdir(<dir-path>)

"""


class PathManagerFactory:
    """
    PathManagerFactory is the class responsible for creating new PathManager
    instances and removing them when no longer needed.
    PathManager can be instantiated directly too, but it is recommended that
    you use PathManagerFactory to create them.
    """

    GLOBAL_PATH_MANAGER = "global_path_manager"
    # pyre-fixme[4]: Attribute must be annotated.
    pm_list = {}

    @staticmethod
    # pyre-fixme[2]: Parameter must be annotated.
    def get(key=GLOBAL_PATH_MANAGER, defaults_setup=False) -> PathManager:
        """
        Get the path manager instance associated with a key.
        A new instance will be created if there is no existing
        instance associated with the key passed in.
        Args:
            key (str):
            defaults_setup (bool): If True, setup_defaults is called.
        """
        if key not in PathManagerFactory.pm_list:
            PathManagerFactory.pm_list[key] = PathManager()
            if defaults_setup:
                try:
                    from iopath.common.setup_defaults import setup_defaults

                    setup_defaults(PathManagerFactory.pm_list[key])
                except ImportError:
                    pass

        return PathManagerFactory.pm_list[key]

    @staticmethod
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def remove(key):
        """
        Remove the path manager instance associated with a key.
        Args:
            key (str):
        """
        if key in PathManagerFactory.pm_list:
            _pm = PathManagerFactory.pm_list.pop(key)  # noqa: F841
            del _pm


"""
A global instance of PathManager.
This global instance is provided for backward compatibility, but it is
recommended that clients use PathManagerFactory
"""
# pyre-fixme[5]:  Global expression must be annotated.
g_pathmgr = PathManagerFactory.get(defaults_setup=True)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from iopath.common.file_io import file_lock, get_cache_dir, LazyPath, PathManager


__all__ = [
    "LazyPath",
    "PathManager",
    "get_cache_dir",
    "file_lock",
]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import os
import shutil
import tempfile
import unittest
import uuid
from contextlib import contextmanager
from typing import Generator, Optional
from unittest.mock import MagicMock, patch

from iopath.common import file_io
from iopath.common.file_io import (
    g_pathmgr,
    get_cache_dir,
    HTTPURLHandler,
    LazyPath,
    OneDrivePathHandler,
    PathManager,
    PathManagerFactory,
)


# from iopath.common.event_logger import EventLogger


# @patch("iopath.common.event_logger.EventLogger.log_event", side_effect=log_event_mock)
class TestNativeIO(unittest.TestCase):

    _tmpdir: Optional[str] = None
    _filename: Optional[str] = None
    _tmpfile: Optional[str] = None
    _tmpfile_contents = "Hello, World"
    _pathmgr = PathManager()

    @classmethod
    def setUpClass(cls) -> None:
        cls._tmpdir = tempfile.mkdtemp()
        cls._filename = "test_file_for_iopath_with_a_really_uncommon_name.txt"
        with open(os.path.join(cls._tmpdir, cls._filename), "w") as f:
            cls._tmpfile = f.name
            f.write(cls._tmpfile_contents)
            f.flush()

    @classmethod
    def tearDownClass(cls) -> None:
        # Cleanup temp working dir.
        if cls._tmpdir is not None:
            shutil.rmtree(cls._tmpdir)  # type: ignore

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def run(self, result=None):
        with patch("iopath.common.event_logger.EventLogger.log_event"):
            super(TestNativeIO, self).run(result)

    def setUp(self) -> None:
        # Reset class variables set by methods before each test.
        self._pathmgr.set_cwd(None)
        self._pathmgr._native_path_handler._non_blocking_io_manager = None
        self._pathmgr._native_path_handler._non_blocking_io_executor = None
        self._pathmgr._async_handlers.clear()

    # @patch("iopath.common.event_logger.EventLogger.log_event")
    # pyre-fixme[3]: Return type must be annotated.
    def log_event_mock(self, topic: Optional[str] = None):
        pass

    def test_open(self) -> None:
        # with patch("iopath.common.event_logger.EventLogger.log_event") as foo:
        # pyre-ignore
        with self._pathmgr.open(self._tmpfile, "r") as f:
            self.assertEqual(f.read(), self._tmpfile_contents)

    def test_factory_open(self) -> None:
        with g_pathmgr.open(self._tmpfile, "r") as f:
            self.assertEqual(f.read(), self._tmpfile_contents)

        _pathmgr = PathManagerFactory.get("test_pm")
        # pyre-fixme[6]: For 1st param expected `str` but got `Optional[str]`.
        with _pathmgr.open(self._tmpfile, "r") as f:
            self.assertEqual(f.read(), self._tmpfile_contents)

        PathManagerFactory.remove("test_pm")

    def test_open_args(self) -> None:
        # with patch("iopath.common.event_logger.EventLogger.log_event") as foo:
        self._pathmgr.set_strict_kwargs_checking(True)
        f = self._pathmgr.open(
            self._tmpfile,  # type: ignore
            mode="r",
            buffering=1,
            encoding="UTF-8",
            errors="ignore",
            newline=None,
            closefd=True,
            opener=None,
        )
        f.close()

    def test_get_local_path(self) -> None:
        self.assertEqual(
            # pyre-ignore
            self._pathmgr.get_local_path(self._tmpfile),
            self._tmpfile,
        )

    def test_get_local_path_forced(self) -> None:
        self.assertEqual(
            # pyre-ignore
            self._pathmgr.get_local_path(self._tmpfile, force=True),
            self._tmpfile,
        )

    def test_exists(self) -> None:
        # pyre-ignore
        self.assertTrue(self._pathmgr.exists(self._tmpfile))
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        fake_path = os.path.join(self._tmpdir, uuid.uuid4().hex)
        self.assertFalse(self._pathmgr.exists(fake_path))

    def test_isfile(self) -> None:
        self.assertTrue(self._pathmgr.isfile(self._tmpfile))  # pyre-ignore
        # This is a directory, not a file, so it should fail
        self.assertFalse(self._pathmgr.isfile(self._tmpdir))  # pyre-ignore
        # This is a non-existing path, so it should fail
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        fake_path = os.path.join(self._tmpdir, uuid.uuid4().hex)
        self.assertFalse(self._pathmgr.isfile(fake_path))

    def test_isdir(self) -> None:
        # pyre-ignore
        self.assertTrue(self._pathmgr.isdir(self._tmpdir))
        # This is a file, not a directory, so it should fail
        # pyre-ignore
        self.assertFalse(self._pathmgr.isdir(self._tmpfile))
        # This is a non-existing path, so it should fail
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        fake_path = os.path.join(self._tmpdir, uuid.uuid4().hex)
        self.assertFalse(self._pathmgr.isdir(fake_path))

    def test_ls(self) -> None:
        # Create some files in the tempdir to ls out.
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        root_dir = os.path.join(self._tmpdir, "ls")
        os.makedirs(root_dir, exist_ok=True)
        files = sorted(["foo.txt", "bar.txt", "baz.txt"])
        for f in files:
            open(os.path.join(root_dir, f), "a").close()

        children = sorted(self._pathmgr.ls(root_dir))
        self.assertListEqual(children, files)

        # Cleanup the tempdir
        shutil.rmtree(root_dir)

    def test_mkdirs(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        new_dir_path = os.path.join(self._tmpdir, "new", "tmp", "dir")
        self.assertFalse(self._pathmgr.exists(new_dir_path))
        self._pathmgr.mkdirs(new_dir_path)
        self.assertTrue(self._pathmgr.exists(new_dir_path))

    def test_copy(self) -> None:
        _tmpfile_2 = self._tmpfile + "2"  # pyre-ignore
        _tmpfile_2_contents = "something else"
        with open(_tmpfile_2, "w") as f:
            f.write(_tmpfile_2_contents)
            f.flush()
        # pyre-fixme[6]: For 1st param expected `str` but got `Optional[str]`.
        self.assertTrue(self._pathmgr.copy(self._tmpfile, _tmpfile_2, overwrite=True))
        with self._pathmgr.open(_tmpfile_2, "r") as f:
            self.assertEqual(f.read(), self._tmpfile_contents)

    def test_move(self) -> None:
        _tmpfile_2 = self._tmpfile + "2" + uuid.uuid4().hex  # pyre-ignore
        _tmpfile_3 = self._tmpfile + "3_" + uuid.uuid4().hex  # pyre-ignore
        _tmpfile_2_contents = "Hello Move"
        with open(_tmpfile_2, "w") as f:
            f.write(_tmpfile_2_contents)
            f.flush()
        self.assertTrue(self._pathmgr.mv(_tmpfile_2, _tmpfile_3))
        with self._pathmgr.open(_tmpfile_3, "r") as f:
            self.assertEqual(f.read(), _tmpfile_2_contents)
        self.assertFalse(self._pathmgr.exists(_tmpfile_2))
        self._pathmgr.rm(_tmpfile_3)

    def test_symlink(self) -> None:
        _symlink = self._tmpfile + "_symlink"  # pyre-ignore
        self.assertTrue(self._pathmgr.symlink(self._tmpfile, _symlink))  # pyre-ignore
        with self._pathmgr.open(_symlink) as f:
            self.assertEqual(f.read(), self._tmpfile_contents)
        self.assertEqual(os.readlink(_symlink), self._tmpfile)
        os.remove(_symlink)

    def test_rm(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        with open(os.path.join(self._tmpdir, "test_rm.txt"), "w") as f:
            rm_file = f.name
            f.write(self._tmpfile_contents)
            f.flush()
        self.assertTrue(self._pathmgr.exists(rm_file))
        self.assertTrue(self._pathmgr.isfile(rm_file))
        self._pathmgr.rm(rm_file)
        self.assertFalse(self._pathmgr.exists(rm_file))
        self.assertFalse(self._pathmgr.isfile(rm_file))

    def test_set_cwd(self) -> None:
        # File not found since cwd not set yet.
        # pyre-fixme[6]: For 1st param expected `str` but got `Optional[str]`.
        self.assertFalse(self._pathmgr.isfile(self._filename))
        # pyre-fixme[6]: For 1st param expected `str` but got `Optional[str]`.
        self.assertTrue(self._pathmgr.isfile(self._tmpfile))
        # Once cwd is set, relative file path works.
        self._pathmgr.set_cwd(self._tmpdir)
        # pyre-fixme[6]: For 1st param expected `str` but got `Optional[str]`.
        self.assertTrue(self._pathmgr.isfile(self._filename))

        # Set cwd to None
        self._pathmgr.set_cwd(None)
        # pyre-fixme[6]: For 1st param expected `str` but got `Optional[str]`.
        self.assertFalse(self._pathmgr.isfile(self._filename))
        # pyre-fixme[6]: For 1st param expected `str` but got `Optional[str]`.
        self.assertTrue(self._pathmgr.isfile(self._tmpfile))

        # Set cwd to invalid path
        with self.assertRaises(ValueError):
            self._pathmgr.set_cwd("/nonexistent/path")

    def test_get_path_with_cwd(self) -> None:
        self._pathmgr.set_cwd(self._tmpdir)
        # Make sure _get_path_with_cwd() returns correctly.
        self.assertEqual(
            # pyre-fixme[6]: For 1st param expected `str` but got `Optional[str]`.
            self._pathmgr._native_path_handler._get_path_with_cwd(self._filename),
            self._tmpfile,
        )
        self.assertEqual(
            self._pathmgr._native_path_handler._get_path_with_cwd("/abs.txt"),
            "/abs.txt",
        )

    def test_bad_args(self) -> None:
        # TODO (T58240718): Replace with dynamic checks
        with self.assertRaises(ValueError):
            self._pathmgr.copy(self._tmpfile, self._tmpfile, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.exists(self._tmpfile, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.get_local_path(self._tmpfile, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.isdir(self._tmpfile, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.isfile(self._tmpfile, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.ls(self._tmpfile, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.mkdirs(self._tmpfile, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.open(self._tmpfile, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.opena(self._tmpfile, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.rm(self._tmpfile, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.set_cwd(self._tmpdir, foo="foo")  # type: ignore

        self._pathmgr.set_strict_kwargs_checking(False)

        self._pathmgr.copy(
            self._tmpfile, self._tmpfile + "2", foo="foo"  # type: ignore
        )
        self._pathmgr.exists(self._tmpfile, foo="foo")  # type: ignore
        self._pathmgr.get_local_path(self._tmpfile, foo="foo")  # type: ignore
        self._pathmgr.isdir(self._tmpfile, foo="foo")  # type: ignore
        self._pathmgr.isfile(self._tmpfile, foo="foo")  # type: ignore
        self._pathmgr.ls(self._tmpdir, foo="foo")  # type: ignore
        self._pathmgr.mkdirs(self._tmpdir, foo="foo")  # type: ignore
        f = self._pathmgr.open(self._tmpfile, foo="foo")  # type: ignore
        f.close()
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        with open(os.path.join(self._tmpdir, "test_rm.txt"), "w") as f:
            rm_file = f.name
            f.write(self._tmpfile_contents)
            f.flush()
        self._pathmgr.rm(rm_file, foo="foo")  # type: ignore


class TestHTTPIO(unittest.TestCase):
    _remote_uri = "https://www.facebook.com"
    _filename = "facebook.html"
    _pathmgr = PathManager()

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def run(self, result=None):
        with patch("iopath.common.event_logger.EventLogger.log_event"):
            super(TestHTTPIO, self).run(result)

    @contextmanager
    def _patch_download(self) -> Generator[None, None, None]:
        def fake_download(url: str, dir: str, *, filename: str) -> str:
            dest = os.path.join(dir, filename)
            with open(dest, "w") as f:
                f.write("test")
            return dest

        with patch.object(
            file_io, "get_cache_dir", return_value=self._cache_dir
        ), patch.object(file_io, "download", side_effect=fake_download):
            yield

    @classmethod
    def setUpClass(cls) -> None:
        # Use a unique pid based cache directory name.
        pid = os.getpid()
        cls._cache_dir: str = os.path.join(get_cache_dir(), f"{__name__}_{pid}")
        cls._pathmgr.register_handler(HTTPURLHandler())
        if os.path.exists(cls._cache_dir):
            shutil.rmtree(cls._cache_dir)
        os.makedirs(cls._cache_dir, exist_ok=True)

    def test_get_local_path(self) -> None:
        with self._patch_download():
            local_path = self._pathmgr.get_local_path(self._remote_uri)
            self.assertTrue(os.path.exists(local_path))
            self.assertTrue(os.path.isfile(local_path))
            # Now get a copy with force set to True.
            local_path = self._pathmgr.get_local_path(self._remote_uri, force=True)
            self.assertTrue(os.path.exists(local_path))
            self.assertTrue(os.path.isfile(local_path))
            # Now get a copy with cache_dir set to a temporary directory.
            local_path = self._pathmgr.get_local_path(
                self._remote_uri, cache_dir=self._cache_dir
            )
            self.assertTrue(local_path.startswith(self._cache_dir))
            self.assertTrue(os.path.exists(local_path))
            self.assertTrue(os.path.isfile(local_path))

    def test_open(self) -> None:
        with self._patch_download():
            with self._pathmgr.open(self._remote_uri, "rb") as f:
                self.assertTrue(os.path.exists(f.name))
                self.assertTrue(os.path.isfile(f.name))
                self.assertTrue(f.read() != "")

    def test_open_writes(self) -> None:
        # HTTPURLHandler does not support writing, only reading.
        with self.assertRaises(AssertionError):
            with self._pathmgr.open(self._remote_uri, "w") as f:
                f.write("foobar")  # pyre-ignore

    def test_open_new_path_manager(self) -> None:
        with self._patch_download():
            path_manager = PathManager()
            with self.assertRaises(OSError):  # no handler registered
                f = path_manager.open(self._remote_uri, "rb")

            path_manager.register_handler(HTTPURLHandler())
            with path_manager.open(self._remote_uri, "rb") as f:
                self.assertTrue(os.path.isfile(f.name))
                self.assertTrue(f.read() != "")

    def test_bad_args(self) -> None:
        with self.assertRaises(NotImplementedError):
            self._pathmgr.copy(
                self._remote_uri, self._remote_uri, foo="foo"  # type: ignore
            )
        with self.assertRaises(NotImplementedError):
            self._pathmgr.exists(self._remote_uri, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.get_local_path(self._remote_uri, foo="foo")  # type: ignore
        with self.assertRaises(NotImplementedError):
            self._pathmgr.isdir(self._remote_uri, foo="foo")  # type: ignore
        with self.assertRaises(NotImplementedError):
            self._pathmgr.isfile(self._remote_uri, foo="foo")  # type: ignore
        with self.assertRaises(NotImplementedError):
            self._pathmgr.ls(self._remote_uri, foo="foo")  # type: ignore
        with self.assertRaises(NotImplementedError):
            self._pathmgr.mkdirs(self._remote_uri, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.open(self._remote_uri, foo="foo")  # type: ignore
        with self.assertRaises(ValueError):
            self._pathmgr.opena(self._remote_uri, foo="foo")  # type: ignore
        with self.assertRaises(NotImplementedError):
            self._pathmgr.rm(self._remote_uri, foo="foo")  # type: ignore
        with self.assertRaises(NotImplementedError):
            self._pathmgr.set_cwd(self._remote_uri, foo="foo")  # type: ignore

        self._pathmgr.set_strict_kwargs_checking(False)

        self._pathmgr.get_local_path(self._remote_uri, foo="foo")  # type: ignore
        f = self._pathmgr.open(self._remote_uri, foo="foo")  # type: ignore
        f.close()
        self._pathmgr.set_strict_kwargs_checking(True)

    def test_copy_across_handlers(self) -> None:
        with tempfile.NamedTemporaryFile(delete=True) as file:
            local_path = file.name
        with self._patch_download():
            self._pathmgr.copy(self._remote_uri, local_path)
            self.assertTrue(self._pathmgr.exists(local_path))

        self._pathmgr.rm(local_path)


class TestLazyPath(unittest.TestCase):
    _pathmgr = PathManager()

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def run(self, result=None):
        with patch("iopath.common.event_logger.EventLogger.log_event"):
            super(TestLazyPath, self).run(result)

    def test_materialize(self) -> None:
        f = MagicMock(return_value="test")
        x = LazyPath(f)
        f.assert_not_called()

        p = os.fspath(x)
        f.assert_called()
        self.assertEqual(p, "test")

        p = os.fspath(x)
        # should only be called once
        f.assert_called_once()
        self.assertEqual(p, "test")

    def test_join(self) -> None:
        f = MagicMock(return_value="test")
        x = LazyPath(f)
        p = os.path.join(x, "a.txt")
        f.assert_called_once()
        self.assertEqual(p, "test/a.txt")

    def test_getattr(self) -> None:
        x = LazyPath(lambda: "abc")
        with self.assertRaises(AttributeError):
            x.startswith("ab")
        _ = os.fspath(x)
        self.assertTrue(x.startswith("ab"))

    def test_PathManager(self) -> None:
        x = LazyPath(lambda: "./")
        output = self._pathmgr.ls(x)  # pyre-ignore
        output_gt = self._pathmgr.ls("./")
        self.assertEqual(sorted(output), sorted(output_gt))

    def test_getitem(self) -> None:
        x = LazyPath(lambda: "abc")
        with self.assertRaises(TypeError):
            x[0]
        _ = os.fspath(x)
        self.assertEqual(x[0], "a")


class TestOneDrive(unittest.TestCase):
    _url = "https://1drv.ms/u/s!Aus8VCZ_C_33gQbJsUPTIj3rQu99"

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def run(self, result=None):
        with patch("iopath.common.event_logger.EventLogger.log_event"):
            super(TestOneDrive, self).run(result)

    def test_one_drive_download(self) -> None:
        _direct_url = OneDrivePathHandler().create_one_drive_direct_download(self._url)
        _gt_url = (
            "https://api.onedrive.com/v1.0/shares/u!aHR0cHM6Ly8xZHJ2Lm1zL3UvcyFBd"
            + "XM4VkNaX0NfMzNnUWJKc1VQVElqM3JRdTk5/root/content"
        )
        self.assertEqual(_direct_url, _gt_url)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import io
import unittest
from unittest.mock import patch

from iopath.common.file_io import PathManager
from iopath.common.s3 import S3ChunkReadIO, S3PathHandler


try:
    import boto3
except ImportError:
    boto3 = None

test_bucket = "TEST_BUCKET_NAME_REPLACE_ME"
test_rel_path = "TEST_REL_PATH_REPLACE_ME"


def test_bucket_defined():
    return (
        test_bucket != "TEST_BUCKET_NAME_REPLACE_ME"
        and test_rel_path != "TEST_REL_PATH_REPLACE_ME"
    )


@unittest.skipIf(not boto3, "Requires boto3 install")
@unittest.skipIf(
    test_bucket == "TEST_BUCKET_NAME_REPLACE_ME", "Test Bucket not specified."
)
@unittest.skipIf(
    test_rel_path == "TEST_REL_PATH_REPLACE_ME", "Test relative path not specified."
)
class TestsS3(unittest.TestCase):
    s3_auth = test_bucket == test_bucket_defined()
    skip_s3_auth_required_tests_message = (
        "Provide an s3 project and bucket you are"
        + "authorised against, then set the s3_auth flag to True"
    )

    def run(self, result=None):
        with patch("iopath.common.event_logger.EventLogger.log_event"):
            super(TestsS3, self).run(result)

    #############################################
    # Shared
    #############################################
    @classmethod
    def setUpClass(cls):
        # NOTE: user should change this location.
        cls.s3_bucket = test_bucket
        # NOTE: user should change this to a valid bucket path that is accessible.
        cls.s3_rel_path = test_rel_path
        cls.s3_full_path = "s3://" + cls.s3_bucket + "/" + cls.s3_rel_path
        cls.s3_pathhandler = S3PathHandler()
        cls.pathmanager = PathManager()
        cls.pathmanager.register_handler(cls.s3_pathhandler)

        # Hack to make the test cases ordered.
        # https://stackoverflow.com/questions/4005695/changing-order-of-unit-tests-in-python
        cls.hack = unittest.TestLoader.sortTestMethodsUsing
        unittest.TestLoader.sortTestMethodsUsing = lambda _, x, y: y > x

    @classmethod
    def tearDownClass(cls):
        # Undo the hack
        unittest.TestLoader.sortTestMethodsUsing = cls.hack

        if not cls.s3_auth:
            return

        # Recursive deletion is not implemented,
        # so let's delete each file and directory.

        # Delete all files
        cls.s3_pathhandler._rm("/".join([cls.s3_full_path, "dir1", "f1_write_string"]))
        cls.s3_pathhandler._rm("/".join([cls.s3_full_path, "dir1", "f2_write_bytes"]))
        cls.s3_pathhandler._rm(
            "/".join([cls.s3_full_path, "dir2", "f1_write_string_from_local"])
        )
        cls.s3_pathhandler._rm(
            "/".join([cls.s3_full_path, "dir2", "f2_write_bytes_from_local"])
        )
        cls.s3_pathhandler._rm(
            "/".join([cls.s3_full_path, "dir2", "f3_write_string_from_local"])
        )
        cls.s3_pathhandler._rm(
            "/".join([cls.s3_full_path, "dir2", "f4_write_bytes_from_local"])
        )
        cls.s3_pathhandler._rm("/".join([cls.s3_full_path, "dir1", "alphabet"]))
        cls.s3_pathhandler._rm("/".join([cls.s3_full_path, "dir1", "large_text"]))

        # Delete all directories.
        cls.s3_pathhandler._rm("/".join([cls.s3_full_path, "dir3", "dir4/"]))
        for i in (1, 2, 3):
            cls.s3_pathhandler._rm("/".join([cls.s3_full_path, f"dir{i}/"]))

        assert cls.s3_pathhandler._ls(cls.s3_full_path) == []

    #############################################
    # Up here, test class attributes,
    # and helpers that don't require S3 access.
    #############################################
    def test_00_supported_prefixes(self):
        supported_prefixes = self.s3_pathhandler._get_supported_prefixes()
        self.assertEqual(supported_prefixes, ["s3://"])

    # # Require S3 Authentication ====>

    #############################################
    # Organization of s3 setup
    # dir1/
    #   f1 <- small (via open)
    #   f2 <- large checkpoint file (via open)
    # dir2/
    #   f3 <- small (via copy(), from dir1)
    #   f4 <- large checkpoint file (via copy_from_local)
    # dir3/
    #   dir4/
    #############################################

    #############################################
    # auth
    # Just check that client loads properly
    #############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_01_add_client_to_handler(self):
        self.s3_pathhandler._get_client(
            "/".join([self.s3_full_path, "path", "file.txt"])
        )
        # self.assertTrue(isinstance(self.s3_pathhandler.client, botocore.client.S3)) # TODO

    # TODO: make sure that the error message displays properly if authentication is messed up.

    #############################################
    # mkdirs
    # Set up the dirs
    # (in BASE)
    #   +dir1
    #   +dir2
    #   +dir3
    #   +dir4
    #############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_02_mkdirs_must_end_with_slash(self):
        with self.assertRaises(AssertionError):
            self.s3_pathhandler._mkdirs("/".join([self.s3_full_path, "fail"]))

    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_03_mkdirs(self):
        # dir{1,2,3} in BASE
        for i in (1, 2, 3):
            self.s3_pathhandler._mkdirs("/".join([self.s3_full_path, f"dir{i}/"]))

        # Make a nested directory in dir3
        self.s3_pathhandler._mkdirs("/".join([self.s3_full_path, "dir3/dir4/"]))

    #############################################
    # open (w/wb)
    #   +f1
    #   +f2
    #############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_04_open_write_mode(self):
        with self.s3_pathhandler._open(
            "/".join([self.s3_full_path, "dir1", "f1_write_string"]), "w"
        ) as f:
            f.write("This is a test of writing a string.")

        with self.s3_pathhandler._open(
            "/".join([self.s3_full_path, "dir1", "f2_write_bytes"]), "wb"
        ) as f:
            f.write(b"This is a test of writing bytes.")

        with self.s3_pathhandler._open(
            "/".join([self.s3_full_path, "dir1", "f1_write_string"]), "w"
        ) as f:
            f.write("This is a test of overwriting a string.")

    #############################################
    # open (r/rb)
    #   read f1
    #   read f2
    #############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_05_open_read_mode(self):
        with self.s3_pathhandler._open(
            "/".join([self.s3_full_path, "dir1", "f1_write_string"]), "r"
        ) as f:
            self.assertEqual(f.read(), "This is a test of overwriting a string.")

        with self.s3_pathhandler._open(
            "/".join([self.s3_full_path, "dir1", "f2_write_bytes"]), "rb"
        ) as f:
            self.assertEqual(f.read(), b"This is a test of writing bytes.")

    #############################################
    # isdir / isfile / exists
    #   test dir{1,2,3,4}
    #   test f{1,2}
    #   test nonexistants
    #############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_06_exists(self):
        # Path does not exist (if file)
        self.assertFalse(
            self.s3_pathhandler._exists("/".join([self.s3_full_path, "dir1", "FAIL"]))
        )
        # Path does not exist (if dir)
        self.assertFalse(
            self.s3_pathhandler._exists("/".join([self.s3_full_path, "FAIL/"]))
        )
        # Path exists (is file)
        self.assertTrue(
            self.s3_pathhandler._exists(
                "/".join([self.s3_full_path, "dir1", "f1_write_string"])
            )
        )
        # Path exists (is dir)
        self.assertTrue(
            self.s3_pathhandler._exists("/".join([self.s3_full_path, "dir1/"]))
        )

    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_07_isdir(self):
        # Path does not exist (if file)
        self.assertFalse(
            self.s3_pathhandler._isdir("/".join([self.s3_full_path, "dir1", "FAIL"]))
        )
        # Path does not exist (if dir)
        self.assertFalse(
            self.s3_pathhandler._isdir("/".join([self.s3_full_path, "FAIL/"]))
        )
        # Path exists (is file)
        self.assertFalse(
            self.s3_pathhandler._isdir(
                "/".join([self.s3_full_path, "dir1", "f1_write_string"])
            )
        )
        # Path exists (is dir)
        self.assertTrue(
            self.s3_pathhandler._isdir("/".join([self.s3_full_path, "dir1/"]))
        )

    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_08_isfile(self):
        # Path does not exist (if file)
        self.assertFalse(
            self.s3_pathhandler._isfile("/".join([self.s3_full_path, "dir1", "FAIL"]))
        )
        # Path does not exist (if dir)
        self.assertFalse(
            self.s3_pathhandler._isfile("/".join([self.s3_full_path, "FAIL/"]))
        )
        # Path exists (is file)
        self.assertTrue(
            self.s3_pathhandler._isfile(
                "/".join([self.s3_full_path, "dir1", "f1_write_string"])
            )
        )
        # Path exists (is dir)
        self.assertFalse(
            self.s3_pathhandler._isfile("/".join([self.s3_full_path, "dir1/"]))
        )

    #############################################
    # copy
    #   copy f1 -> f3
    #############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_09_copy(self):
        self.assertTrue(
            self.s3_pathhandler._copy(
                "/".join([self.s3_full_path, "dir1", "f1_write_string"]),
                "/".join([self.s3_full_path, "dir2", "f3_write_string"]),
            )
        )

    #############################################
    # ls
    #   ls dir{1,2,3,4}
    #############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_10_ls(self):
        # Path does not exist (if file)
        self.assertEqual(
            [], self.s3_pathhandler._ls("/".join([self.s3_full_path, "dir1", "FAIL"]))
        )
        # Path does not exist (if dir)
        self.assertEqual(
            [], self.s3_pathhandler._ls("/".join([self.s3_full_path, "FAIL/"]))
        )
        # Path exists (is file)
        self.assertEqual(
            ["/".join([self.s3_rel_path, "dir1", "f1_write_string"])],
            self.s3_pathhandler._ls(
                "/".join([self.s3_full_path, "dir1", "f1_write_string"])
            ),
        )
        # Path exists (is dir)
        self.assertEqual(
            {
                "/".join([self.s3_rel_path, "dir1/"]),
                "/".join([self.s3_rel_path, "dir1", "f1_write_string"]),
                "/".join([self.s3_rel_path, "dir1", "f2_write_bytes"]),
            },
            set(self.s3_pathhandler._ls("/".join([self.s3_full_path, "dir1/"]))),
        )

    #############################################
    # rm
    #   rm f3
    #############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_11_rm(self):
        path = "/".join([self.s3_full_path, "dir2", "f3_write_string"])
        self.assertTrue(self.s3_pathhandler._exists(path))
        self.assertTrue(self.s3_pathhandler._isfile(path))
        self.assertFalse(self.s3_pathhandler._isdir(path))

        self.s3_pathhandler._rm(path)

        self.assertFalse(self.s3_pathhandler._exists(path))
        self.assertFalse(self.s3_pathhandler._isfile(path))
        self.assertFalse(self.s3_pathhandler._isdir(path))

    #############################################
    # get_local_path
    #   Retrieve f{1,2}
    #   Check file contents.
    #############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_12_get_local_path(self):
        s3_path_f1 = "/".join([self.s3_full_path, "dir1", "f1_write_string"])
        s3_path_f2 = "/".join([self.s3_full_path, "dir1", "f2_write_bytes"])

        local_path_f1 = self.s3_pathhandler._get_local_path(s3_path_f1)
        local_path_f2 = self.s3_pathhandler._get_local_path(s3_path_f2)

        with open(local_path_f1, "r") as f:
            self.assertEqual(f.read(), "This is a test of overwriting a string.")
        with open(local_path_f2, "rb") as f:
            self.assertEqual(f.read(), b"This is a test of writing bytes.")

    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_13_get_local_path_idempotent(self):
        """
        Call _get_local_path multiple times.
        Check that we keep returning the same cached copy instead of redownloading.
        """
        s3_path_f1 = "/".join([self.s3_full_path, "dir1", "f1_write_string"])

        REPEATS = 3
        local_paths = [
            self.s3_pathhandler._get_local_path(s3_path_f1) for _ in range(REPEATS)
        ]
        for local_path in local_paths[1:]:
            self.assertEqual(local_path, local_paths[0])

        with open(local_paths[0], "r") as f:
            self.assertEqual(f.read(), "This is a test of overwriting a string.")

    ##############################################
    # copy_from_local
    #   Upload local copies of f1, f2 -> f3, f4.
    # Check contents via open(), and via another get_local_path
    ##############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_14_copy_from_local(self):
        s3_src_path_f1 = "/".join([self.s3_full_path, "dir1", "f1_write_string"])
        s3_src_path_f2 = "/".join([self.s3_full_path, "dir1", "f2_write_bytes"])

        local_path_f1 = self.s3_pathhandler._get_local_path(s3_src_path_f1)
        local_path_f2 = self.s3_pathhandler._get_local_path(s3_src_path_f2)

        s3_dst_path_f1 = "/".join(
            [self.s3_full_path, "dir2", "f1_write_string_from_local"]
        )
        s3_dst_path_f2 = "/".join(
            [self.s3_full_path, "dir2", "f2_write_bytes_from_local"]
        )

        self.assertTrue(
            self.s3_pathhandler._copy_from_local(local_path_f1, s3_dst_path_f1)
        )
        self.assertTrue(
            self.s3_pathhandler._copy_from_local(local_path_f2, s3_dst_path_f2)
        )

    #############################################
    # symlink
    #   should fail
    #############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_15_symlink(self):
        s3_src_path_f1 = "/".join([self.s3_full_path, "dir1", "f1_write_string"])
        s3_dst_path_f1 = "/".join(
            [self.s3_full_path, "dir2", "f1_write_string_symlink"]
        )
        with self.assertRaises(NotImplementedError):
            self.s3_pathhandler._symlink(s3_src_path_f1, s3_dst_path_f1)

    #############################################
    # PathManager
    #   Minimal ls() test with PathManager
    #############################################
    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_16_pathmanager(self):
        # Path exists (is file)
        self.assertEqual(
            ["/".join([self.s3_rel_path, "dir1", "f1_write_string"])],
            self.pathmanager.ls(
                "/".join([self.s3_full_path, "dir1", "f1_write_string"])
            ),
        )

    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_17_chunk_reading(self):
        with self.s3_pathhandler._open(
            "/".join([self.s3_full_path, "dir1", "alphabet"]), "w"
        ) as f:
            f.write("abcdefghijklmnopqrstuvwxyz")

        session = boto3.Session()
        client = session.client("s3")

        reader = S3ChunkReadIO(
            client,
            bucket=self.s3_bucket,
            key="/".join([self.s3_rel_path, "dir1", "alphabet"]),
            chunk_size=6,
        )

        reader.seek(2)

        self.assertEqual(reader.read(4).decode("utf-8"), "cdef")
        self.assertEqual(reader.read(13).decode("utf-8"), "ghijklmnopqrs")
        self.assertEqual(reader.read().decode("utf-8"), "tuvwxyz")

    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_19_test_chunk_read_text(self):
        large_text = "HELLO!\nTHIS IS A TEST\nTHANKS!" * 10 * 1024
        with self.s3_pathhandler._open(
            "/".join([self.s3_full_path, "dir1", "large_text"]), "w"
        ) as wf:
            wf.write(large_text)
            wf.close()

        for test_chunk_size in [4 * 1024, 8 * 1024, 12 * 1024, 16 * 1024, 1024 * 1024]:
            with self.s3_pathhandler._open(
                "/".join([self.s3_full_path, "dir1", "large_text"]),
                "r",
                read_chunk_size=test_chunk_size,
            ) as rf:
                delimiter = "\n"
                expected_lines = large_text.split(delimiter)[:-1]
                for line in expected_lines:
                    s = rf.readline()
                    self.maxDiff = None
                    self.assertEqual(s, line + delimiter)
                rf.close()

    @unittest.skipIf(not s3_auth, skip_s3_auth_required_tests_message)
    def test_20_test_chunk_read_binary(self):
        large_text = "HELLO!\nTHIS IS A TEST\nTHANKS!" * 10 * 1024
        with self.s3_pathhandler._open(
            "/".join([self.s3_full_path, "dir1", "large_text_binary"]), "wb"
        ) as wf:
            wf.write(large_text.encode())
            wf.close()

        for test_chunk_size in [4 * 1024, 8 * 1024, 12 * 1024, 16 * 1024, 1024 * 1024]:
            buffer = io.BytesIO()
            with self.s3_pathhandler._open(
                "/".join([self.s3_full_path, "dir1", "large_text_binary"]),
                "rb",
                read_chunk_size=test_chunk_size,
            ) as rf:
                while True:
                    chunk = rf.read(6 * 1024)
                    if len(chunk) > 0:
                        buffer.write(chunk)
                    else:
                        break
                self.assertEqual(large_text.encode(), buffer.getvalue())
                rf.close()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import os
import unittest
import uuid
from unittest.mock import patch

from iopath.common.download import download


class TestDownload(unittest.TestCase):
    # pyre-fixme[4]: Attribute must be annotated.
    _filename = "facebook_" + uuid.uuid4().hex + ".html"

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def run(self, result=None):
        with patch("iopath.common.event_logger.EventLogger.log_event"):
            super(TestDownload, self).run(result)

    def test_download(self) -> None:
        download(
            "https://www.facebook.com", ".", filename=self._filename, progress=False
        )
        self.assertTrue(os.path.isfile(self._filename))

    def tearDown(self) -> None:
        if os.path.isfile(self._filename):
            os.unlink(self._filename)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import Dict, Optional, TypeVar


try:
    from tmetry.simpleevent import SimpleEventRecord
    from tmetry.writer import TmetryWriter

    b_tmetry_available = True
except ImportError:
    b_tmetry_available = False

VTYPE = TypeVar("T", str, int, bool, float)


class EventLogger:
    """
    Base class for providing event logging in a path handler
    It implements event logging by wrapping the tmetry interface.
    If tmetry packages is not available, it is a no-op.
    """

    DEFAULT_TOPIC = "iopath_tmetry"
    OP_KEY = "op"

    # Interval after which event is logged to backend.
    SAMPLING_PERIOD = 10

    # Map to keep track of sample count per operation.
    # pyre-fixme[4]: Attribute must be annotated.
    sample_counts = {}

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def __init__(self, *args, **kwargs):
        if b_tmetry_available:
            # pyre-fixme[4]: Attribute must be annotated.
            self._writers = []
            # pyre-fixme[4]: Attribute must be annotated.
            self._evt = SimpleEventRecord()
            # pyre-fixme[4]: Attribute must be annotated.
            self._enabled = True

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def add_writer(self, writer):
        if b_tmetry_available:
            if isinstance(writer, TmetryWriter):
                self._writers.append(writer)

    # pyre-fixme[3]: Return type must be annotated.
    def add_key(self, key: str, val: VTYPE):
        if b_tmetry_available:
            self._evt.set(key, val)

    # pyre-fixme[3]: Return type must be annotated.
    def add_keys(self, kvs: Dict[str, VTYPE]):
        if b_tmetry_available:
            self._evt.set_keys(kvs)

    def _sample_record(self) -> bool:
        """
        Samples the current event and logs only when the count
        reaches logging interval.

        Returns:
            True: if this sample should be logged.
            False: otherwise.
        """
        evt_op = self._evt.get(self.OP_KEY)
        if evt_op is None:
            # No op is set. Let's log it.
            return True

        if evt_op not in self.sample_counts:
            self.sample_counts[evt_op] = 1
            return True

        self.sample_counts[evt_op] += 1
        if self.sample_counts[evt_op] > self.SAMPLING_PERIOD:
            # Let's log this and reset sanpling counter.
            self.sample_counts[evt_op] = 1
            return True

        # Skip this sample.
        return False

    def set_logging(self, enable: bool) -> None:
        self._enabled = enable

    def is_logging_enabled(self) -> bool:
        return self._enabled

    # pyre-fixme[3]: Return type must be annotated.
    def log_event(self, topic: Optional[str] = None):
        if b_tmetry_available and self._enabled:

            # Sample the current event.
            if not self._sample_record():
                return

            if topic is None:
                topic = self.DEFAULT_TOPIC

            for writer in self._writers:
                writer.writeRecord(topic, self._evt)
            del self._evt
            self._evt = SimpleEventRecord()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import concurrent.futures
import io
import os
import shutil
import tempfile
import unittest
from typing import Optional
from unittest.mock import Mock, patch

from iopath.common.file_io import NativePathHandler, PathManager
from iopath.common.non_blocking_io import (
    NonBlockingBufferedIO,
    NonBlockingIO,
    NonBlockingIOManager,
)


class TestNativeIOAsync(unittest.TestCase):
    """
    This test class is meant to have comprehensive tests for
    `NativePathHandler`. Async functionality tests for other
    `PathHandler`-s should only require a single test since
    all `PathHandler`-s operate in the same way.
    """

    _tmpdir: Optional[str] = None
    _pathmgr = PathManager()

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def run(self, result=None):
        with patch("iopath.common.event_logger.EventLogger.log_event"):
            super(TestNativeIOAsync, self).run(result)

    @classmethod
    def setUpClass(cls) -> None:
        cls._tmpdir = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls) -> None:
        # Cleanup temp working dir.
        if cls._tmpdir is not None:
            shutil.rmtree(cls._tmpdir)  # type: ignore

    def setUp(self) -> None:
        # Reset class variables set by methods before each test.
        self._pathmgr.set_cwd(None)
        self._pathmgr._native_path_handler._non_blocking_io_manager = None
        self._pathmgr._native_path_handler._non_blocking_io_executor = None
        self._pathmgr._async_handlers.clear()

    def test_opena(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _tmpfile = os.path.join(self._tmpdir, "async.txt")
        try:
            # Write the files.
            with self._pathmgr.opena(_tmpfile + "f", "w") as f:
                f.write("f1 ")
                with self._pathmgr.opena(_tmpfile + "g", "w") as g:
                    f.write("f2 ")
                    g.write("g1 ")
                    f.write("f3 ")
                f.write("f4 ")
            with self._pathmgr.opena(_tmpfile + "f", "a") as f:
                f.write("f5 ")
            F_STR = "f1 f2 f3 f4 f5 "
            G_STR = "g1 "

            # Test that `PathManager._async_handlers` keeps track of all
            # `PathHandler`-s where `opena` is used.
            self.assertCountEqual(
                [type(handler) for handler in self._pathmgr._async_handlers],
                [type(self._pathmgr._native_path_handler)],
            )
            # Test that 2 paths were properly logged in `NonBlockingIOManager`.
            manager = self._pathmgr._native_path_handler._non_blocking_io_manager
            self.assertEqual(len(manager._path_to_data), 2)
        finally:
            # Join the threads to wait for files to be written.
            self.assertTrue(self._pathmgr.async_close())

        # Check that both files were asynchronously written and written in order.
        with self._pathmgr.open(_tmpfile + "f", "r") as f:
            self.assertEqual(f.read(), F_STR)
        with self._pathmgr.open(_tmpfile + "g", "r") as g:
            self.assertEqual(g.read(), G_STR)
        # Test that both `NonBlockingIO` objects `f` and `g` are finally closed.
        self.assertEqual(len(manager._path_to_data), 0)

    def test_async_join_behavior(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _tmpfile = os.path.join(self._tmpdir, "async.txt")
        _tmpfile_contents = "Async Text"
        try:
            for _ in range(1):  # Opens 1 thread
                with self._pathmgr.opena(_tmpfile + "1", "w") as f:
                    f.write(f"{_tmpfile_contents}-1")
            for _ in range(2):  # Opens 2 threads
                with self._pathmgr.opena(_tmpfile + "2", "w") as f:
                    f.write(f"{_tmpfile_contents}-2")
            for _ in range(3):  # Opens 3 threads
                with self._pathmgr.opena(_tmpfile + "3", "w") as f:
                    f.write(f"{_tmpfile_contents}-3")
            _path_to_data = (
                self._pathmgr._native_path_handler._non_blocking_io_manager._path_to_data
            )
            # Join the threads for the 1st and 3rd file and ensure threadpool completed.
            _path_to_data_copy = dict(_path_to_data)
            self.assertTrue(
                self._pathmgr.async_join(
                    _tmpfile + "1", _tmpfile + "3"
                )  # Removes paths from `_path_to_io`.
            )
            self.assertFalse(_path_to_data_copy[_tmpfile + "1"].thread.is_alive())
            self.assertFalse(_path_to_data_copy[_tmpfile + "3"].thread.is_alive())
            self.assertEqual(len(_path_to_data), 1)  # 1 file remaining
        finally:
            # Join all the remaining threads
            _path_to_data_copy = dict(_path_to_data)
            self.assertTrue(self._pathmgr.async_close())

        # Ensure data cleaned up.
        self.assertFalse(_path_to_data_copy[_tmpfile + "2"].thread.is_alive())
        self.assertEqual(len(self._pathmgr._async_handlers), 0)
        self.assertEqual(len(_path_to_data), 0)  # 0 files remaining

    def test_opena_normpath(self) -> None:
        _filename = "async.txt"
        # `_file1` and `_file2` should represent the same path but have different
        # string representations.
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file1 = os.path.join(self._tmpdir, _filename)
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file2 = os.path.join(self._tmpdir, ".", _filename)
        self.assertNotEqual(_file1, _file2)
        try:
            _file1_text = "File1 text"
            _file2_text = "File2 text"
            with self._pathmgr.opena(_file1, "w") as f:
                f.write(_file1_text)
            with self._pathmgr.opena(_file2, "a") as f:
                f.write(_file2_text)
            _path_to_data = (
                self._pathmgr._native_path_handler._non_blocking_io_manager._path_to_data
            )
            # Check that `file2` is marked as the same file as `file1`.
            self.assertEqual(len(_path_to_data), 1)
            self.assertTrue(self._pathmgr.async_join())
            # Check that both file paths give the same file contents.
            with self._pathmgr.open(_file1, "r") as f:
                self.assertEqual(f.read(), _file1_text + _file2_text)
            with self._pathmgr.open(_file2, "r") as f:
                self.assertEqual(f.read(), _file1_text + _file2_text)
        finally:
            self.assertTrue(self._pathmgr.async_close())

    def test_async_consecutive_join_calls(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file = os.path.join(self._tmpdir, "async.txt")
        try:
            self.assertTrue(self._pathmgr.async_join())
            try:
                with self._pathmgr.opena(_file, "w") as f:
                    f.write("1")
            finally:
                self.assertTrue(self._pathmgr.async_join())
            with self._pathmgr.open(_file, "r") as f:
                self.assertEqual(f.read(), "1")

            try:
                f = self._pathmgr.opena(_file, "a")
                f.write("2")
                f.close()
            finally:
                self.assertTrue(self._pathmgr.async_join())
            with self._pathmgr.open(_file, "r") as f:
                self.assertEqual(f.read(), "12")
        finally:
            self.assertTrue(self._pathmgr.async_close())

    def test_opena_mode_restriction(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file = os.path.join(self._tmpdir, "async.txt")
        with self.assertRaises(ValueError):
            self._pathmgr.opena(_file, "r")
        with self.assertRaises(ValueError):
            self._pathmgr.opena(_file, "rb")
        with self.assertRaises(ValueError):
            self._pathmgr.opena(_file, "wrb")

    def test_opena_args_passed_correctly(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file = os.path.join(self._tmpdir, "async.txt")
        try:
            # Make sure that `opena` args are used correctly by using
            # different newline args.
            with self._pathmgr.opena(_file, "w", newline="\r\n") as f:
                f.write("1\n")
            with self._pathmgr.opena(_file, "a", newline="\n") as f:
                f.write("2\n3")
        finally:
            self.assertTrue(self._pathmgr.async_close())

        # Read the raw file data without converting newline endings to see
        # if the `opena` args were used correctly.
        with self._pathmgr.open(_file, "r", newline="") as f:
            self.assertEqual(f.read(), "1\r\n2\n3")

    def test_opena_with_callback(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file_tmp = os.path.join(self._tmpdir, "async.txt.tmp")
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file = os.path.join(self._tmpdir, "async.txt")
        _data = "Asynchronously written text"

        # pyre-fixme[53]: Captured variable `_data` is not annotated.
        # pyre-fixme[53]: Captured variable `_file` is not annotated.
        # pyre-fixme[53]: Captured variable `_file_tmp` is not annotated.
        # pyre-fixme[3]: Return type must be annotated.
        def cb():
            # Insert a test to make sure `_file_tmp` was closed before
            # the callback is called.
            with open(_file_tmp, "r") as f:
                self.assertEqual(f.read(), _data)
            self._pathmgr.copy(_file_tmp, _file, overwrite=True)

        mock_cb = Mock(side_effect=cb)

        try:
            with self._pathmgr.opena(
                _file_tmp, "w", callback_after_file_close=mock_cb
            ) as f:
                f.write(_data)
        finally:
            self.assertTrue(self._pathmgr.async_close())
        # Callback should have been called exactly once.
        mock_cb.assert_called_once()

        # Data should have been written to both `_file_tmp` and `_file`.
        with open(_file_tmp, "r") as f:
            self.assertEqual(f.read(), _data)
        with open(_file, "r") as f:
            self.assertEqual(f.read(), _data)

    def test_opena_with_callback_only_called_once(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file_tmp = os.path.join(self._tmpdir, "async.txt.tmp")

        mock_cb = Mock()

        # Callback should be called once even if `close` is called
        # multiple times.
        try:
            f = self._pathmgr.opena(_file_tmp, "w", callback_after_file_close=mock_cb)
            f.close()
            f.close()
            f.close()
        finally:
            self.assertTrue(self._pathmgr.async_close())
        # Callback should have been called exactly once.
        mock_cb.assert_called_once()

    def test_async_custom_executor(self) -> None:
        # At first, neither manager nor executor are set.
        self.assertIsNone(self._pathmgr._native_path_handler._non_blocking_io_manager)
        self.assertIsNone(self._pathmgr._native_path_handler._non_blocking_io_executor)
        # Then, override the `NativePathHandler` and set a custom executor.
        executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=128, thread_name_prefix="my prefix"
        )
        ph = NativePathHandler(async_executor=executor)
        self._pathmgr.register_handler(ph, allow_override=True)
        self.assertEqual(ph, self._pathmgr._native_path_handler)

        # Opening a file with `opena` initializes the manager with the
        # executor.
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file = os.path.join(self._tmpdir, "async.txt")
        try:
            with self._pathmgr.opena(_file, "w") as f:
                f.write("Text")
            # Make sure the manager's executor is the same as the user's.
            self.assertEqual(
                executor,
                self._pathmgr._native_path_handler._non_blocking_io_manager._pool,
            )
        finally:
            self.assertTrue(self._pathmgr.async_close())

    def test_non_blocking_io_seekable(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file = os.path.join(self._tmpdir, "async.txt")
        # '^' marks the current position in stream

        # Test seek.
        try:
            with self._pathmgr.opena(_file, "wb") as f:
                f.write(b"012345")  # file = 012345^
                f.seek(1)  # file = 0^12345
                f.write(b"##")  # file = 0##^345
        finally:
            self.assertTrue(self._pathmgr.async_join())
            with self._pathmgr.open(_file, "rb") as f:
                self.assertEqual(f.read(), b"0##345")

        # Test truncate.
        try:
            with self._pathmgr.opena(_file, "wb") as f:
                f.write(b"012345")  # file = 012345^
                f.seek(2)  # file = 01^2345
                f.truncate()  # file = 01^
        finally:
            self.assertTrue(self._pathmgr.async_join())
            with self._pathmgr.open(_file, "rb") as f:
                self.assertEqual(f.read(), b"01")

        # Big test for seek and truncate.
        try:
            with self._pathmgr.opena(_file, "wb") as f:
                f.write(b"0123456789")  # file = 0123456789^
                f.seek(2)  # file = 01^23456789
                f.write(b"##")  # file = 01##^456789
                f.seek(3, io.SEEK_CUR)  # file = 01##456^789
                f.truncate()  # file = 01##456^
                f.write(b"$")  # file = 01##456$^
        finally:
            self.assertTrue(self._pathmgr.async_join())
            with self._pathmgr.open(_file, "rb") as f:
                self.assertEqual(f.read(), b"01##456$")

        # Test NOT tellable.
        try:
            with self._pathmgr.opena(_file, "wb") as f:
                with self.assertRaises(ValueError):
                    f.tell()
        finally:
            self._pathmgr.async_close()


class TestNonBlockingIO(unittest.TestCase):
    _tmpdir: Optional[str] = None
    _io_manager = NonBlockingIOManager(buffered=False)
    _buffered_io_manager = NonBlockingIOManager(buffered=True)

    @classmethod
    def setUpClass(cls) -> None:
        cls._tmpdir = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls) -> None:
        # Cleanup temp working dir.
        if cls._tmpdir is not None:
            shutil.rmtree(cls._tmpdir)  # type: ignore

    def test_select_io(self) -> None:
        self.assertEqual(self._io_manager._IO, NonBlockingIO)
        self.assertEqual(self._buffered_io_manager._IO, NonBlockingBufferedIO)

    def test_io_manager(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file = os.path.join(self._tmpdir, "non_buffered.txt")

        try:
            # Test IO notifies manager after every write call.
            f = self._io_manager.get_non_blocking_io(
                path=_file, io_obj=open(_file, "w")
            )
            with patch.object(
                f,
                "_notify_manager",
                # pyre-fixme[16]: Item `IO` of `Union[IO[bytes], IO[str]]` has no
                #  attribute `_notify_manager`.
                wraps=f._notify_manager,
            ) as mock_notify_manager:
                f.write("." * 1)
                f.write("." * 2)
                f.write("." * 3)
                # Should notify manager 3 times: 3 write calls.
                self.assertEqual(mock_notify_manager.call_count, 3)
                mock_notify_manager.reset_mock()
                # Should notify manager 1 time: 1 close call.
                f.close()
                self.assertEqual(mock_notify_manager.call_count, 1)
        finally:
            self.assertTrue(self._io_manager._join())
            self.assertTrue(self._io_manager._close_thread_pool())

        with open(_file, "r") as f:
            self.assertEqual(f.read(), "." * 6)

    def test_buffered_io_manager(self) -> None:
        # pyre-fixme[6]: For 1st param expected `Union[PathLike[str], str]` but got
        #  `Optional[str]`.
        _file = os.path.join(self._tmpdir, "buffered.txt")

        try:
            # Test IO doesn't flush until buffer is full.
            f = self._buffered_io_manager.get_non_blocking_io(
                path=_file, io_obj=open(_file, "wb"), buffering=10
            )
            with patch.object(f, "flush", wraps=f.flush) as mock_flush:
                with patch.object(
                    f,
                    "_notify_manager",
                    # pyre-fixme[16]: Item `IO` of `Union[IO[bytes], IO[str]]` has
                    #  no attribute `_notify_manager`.
                    wraps=f._notify_manager,
                ) as mock_notify_manager:
                    f.write(b"." * 9)
                    mock_flush.assert_not_called()  # buffer not filled - don't flush
                    mock_notify_manager.assert_not_called()
                    # Should flush when full.
                    f.write(b"." * 13)
                    mock_flush.assert_called_once()  # buffer filled - should flush
                    # `flush` should notify manager 4 times: 3 `file.write` and 1 `buffer.close`.
                    # Buffer is split into 3 chunks of size 10, 10, and 2.
                    # pyre-fixme[16]: Item `IO` of `Union[IO[bytes], IO[str]]` has
                    #  no attribute `_buffers`.
                    self.assertEqual(len(f._buffers), 2)  # 22-byte and 0-byte buffers
                    self.assertEqual(mock_notify_manager.call_count, 4)
                    mock_notify_manager.reset_mock()
                    # `close` should notify manager 2 times: 1 `buffer.close` and 1 `file.close`.
                    f.close()
                    self.assertEqual(mock_notify_manager.call_count, 2)

            # Test IO flushes on file close.
            f = self._buffered_io_manager.get_non_blocking_io(
                path=_file, io_obj=open(_file, "ab"), buffering=10
            )
            with patch.object(f, "flush", wraps=f.flush) as mock_flush:
                f.write(b"." * 5)
                mock_flush.assert_not_called()
                f.close()
                mock_flush.assert_called()  # flush on exit
        finally:
            self.assertTrue(self._buffered_io_manager._join())
            self.assertTrue(self._buffered_io_manager._close_thread_pool())

        with open(_file, "rb") as f:
            self.assertEqual(f.read(), b"." * 27)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from iopath.common import file_lock, get_cache_dir, LazyPath, PathManager
from iopath.tabular.tabular_io import TabularPathHandler, TabularUriParser

from .version import __version__


# pyre-fixme[5]: Global expression must be annotated.
__all__ = [
    "LazyPath",
    "PathManager",
    "get_cache_dir",
    "file_lock",
    "TabularPathHandler",
    "TabularUriParser",
    __version__,
]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
    Utility for testing async writes with `torch.save`.
    Usage:
        buck run @mode/opt //fair_infra/data/iopath/tests:async_torch_test
"""

import os
import tempfile
import time

import torch
import torch.nn as nn
import torch.optim as optim
from iopath.common.file_io import PathManager


class Model(nn.Module):
    # pyre-fixme[3]: Return type must be annotated.
    def __init__(self):
        super(Model, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)


class TestDriver:
    _pathmgr = PathManager()

    def test(self) -> None:
        model = Model()
        optimizer = optim.SGD(model.parameters(), lr=0.001, momentum=0.9)

        print("Model's state_dict:")
        for param_tensor in model.state_dict():
            print(f"{param_tensor}\t{model.state_dict()[param_tensor].size()}")
        print("Optimizer's state_dict:")
        for var_name in optimizer.state_dict():
            print(f"{var_name}\t{optimizer.state_dict()[var_name]}")

        with tempfile.TemporaryDirectory() as _tmpdir:
            try:
                URI = os.path.join(_tmpdir, "test.ckpt")

                f = self._pathmgr.opena(URI, "wb")
                i = "*"
                large = f"{i}" * 1000000000

                print("Starting `torch.save` call.")
                torch.save(
                    {
                        "model_state_dict": model.state_dict(),
                        "optimizer_state_dict": optimizer.state_dict(),
                        "large": large,
                    },
                    # pyre-fixme[6]: For 2nd param expected
                    #  `Union[PathLike[typing.Any], IO[bytes], str, BinaryIO]` but got
                    #  `IOBase`.
                    f,
                )
                f.close()
                start_time = time.time()

            finally:
                # We want this `join` call to take time. If it is instantaneous,
                # then our async write calls are not running asynchronously.
                print("Waiting for `torch.save` call to complete at `async_join()`.")
                self._pathmgr.async_join()

            print(
                "Time Python waited for `async_join()` call to finish: "
                f"{time.time() - start_time}s."
            )
            assert self._pathmgr.async_close()

            checkpoint = torch.load(URI)
            for key_item_1, key_item_2 in zip(
                model.state_dict().items(), checkpoint["model_state_dict"].items()
            ):
                assert torch.equal(key_item_1[1], key_item_2[1])
            assert optimizer.state_dict() == checkpoint["optimizer_state_dict"]
            assert large == checkpoint["large"]

            print("Async `torch.save` Test succeeded.")


if __name__ == "__main__":
    print("Async `torch.save` Test starting.")
    tst = TestDriver()
    tst.test()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import datetime as dt
import io
import logging
import os
import shutil
import types
from datetime import datetime, timedelta
from functools import partial
from typing import Any, Dict, IO, List, Optional, Tuple, Union

from iopath.common.file_io import file_lock, get_cache_dir, PathHandler


try:
    # Needed for S3 PathHandler
    # pyre-fixme[21]: Could not find module `boto3`.
    import boto3

    # pyre-fixme[21]: Could not find module `botocore`.
    import botocore
except ImportError:
    boto3 = None
    botocore = None


# Override for close() on files to write to Amazon S3
# pyre-fixme[2]: Parameter must be annotated.
def s3_close_and_upload(self, client, bucket, s3_path, transfer_config) -> None:
    # Seek to start, for use by upload_fileobj.
    self.seek(0)

    # Reinstall the proper close.
    self.close = self._close

    # upload_fileobj needs bytes
    # NOTE: This is undesirable.
    if isinstance(self, io.StringIO):
        self = io.BytesIO(self.getvalue().encode("utf-8"))

    # Upload
    try:
        client.upload_fileobj(
            self,
            bucket,
            s3_path,
            Config=transfer_config,
        )
    except botocore.exceptions.ClientError as e:
        raise OSError(f"Error in file upload - {e}" f"{type(e).__name__}: {e}") from e


class S3PathHandler(PathHandler):
    """
    Support for Amazon Simple Storage Service (S3)

    PathHanlder methods, at a glance:

     File     --torch.load->     In     --open(..., 'w')->   Amazon    <- _exists,_isfile,_isdir,_ls,_rm ...
    System   <-torch.save--     Mem.   <-open(..., 'r')--      S3
            <----------------_copy_from_local-----------------
            ----------------_get_local_path ----------------->

    Mem usage, for processing N bytes:
        open(..., mode)
            mode=='w':    2N,  due to fully buffering user input,
                                *and doing naive conversion from StringIO -> BytesIO*,
                                before writing to S3
                                ^ Potential for optimization.
            mode=='wb':    N,  due to fully buffering user input, before writing to S3.
            mode=='r':     N,  due to fully buffering file in memory
            mode=='rb':    N,  due to fully buffering file in memory
        _copy_from_local: ≈0.  boto3 streams from file system directly to s3
        _get_local_path:  ≈0.  boto3 streams from s3 directly from s3 to file system

    NOTE:
        S3 doesn't have a notion of directories.  This pathhandler simulates
        directories via uploading objects with a name ending in a slash, on calls to mkdir().
        ls() calls return objects as if they were in a directory structure, via
        boto3's options.
    """

    # Disable failures if not all args are specified.
    _strict_kwargs_check = False

    S3_PREFIX = "s3://"
    CACHE_SUBDIR_NAME = "s3_cache"

    # pyre-fixme[3]: Return type must be annotated.
    def __init__(
        self,
        cache_dir: Optional[str] = None,
        profile: Optional[str] = "saml",
        # pyre-fixme[24]: Generic type `dict` expects 2 type parameters, use
        #  `typing.Dict` to avoid runtime subscripting errors.
        transfer_config_kwargs: Optional[Dict] = None,
    ):
        """
        Args:
            cache_dir (str): Local filesystem directory to use for caching. If None,
                uses default from `file_io.get_cache_dir()`.
            transfer_config_kwargs (dict): Settings for boto3.s3.transfer.TransferConfig.
                Used to specify settings for multipart transfers.
                See https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3.html for details.
        """
        self.cache_dir = cache_dir
        self.profile = profile
        # pyre-fixme[21]: Could not find module `boto3.s3.transfer`.
        from boto3.s3.transfer import TransferConfig

        # pyre-fixme[4]: Attribute must be annotated.
        self.transfer_config = TransferConfig(
            **(transfer_config_kwargs if transfer_config_kwargs else {})
        )

    def _get_supported_prefixes(self) -> List[str]:
        """
        Returns:
            List[str]: the list of URI prefixes this PathHandler can support
        """
        return [self.S3_PREFIX]

    def _parse_uri(self, uri: str) -> Tuple[str, str]:
        """
        Parses a "s3://bucket/path" URI into `bucket` and `path` strings.

        Args:
            uri (str): A s3:// URI.

        Returns:
            bucket (str): the s3 bucket.
            path (str): the path on the s3 system.
        """
        splits = uri.replace(self.S3_PREFIX, "").split("/")
        bucket = splits[0]
        path = "/".join(splits[1:])
        return bucket, path

    # pyre-fixme[3]: Return type must be annotated.
    def _get_client(self, bucket: str):
        logger = logging.getLogger(__name__)
        if not hasattr(self, "client"):
            try:
                session = boto3.Session(profile_name=self.profile)
                # pyre-fixme[16]: `S3PathHandler` has no attribute `client`.
                self.client = session.client("s3")
            except botocore.exceptions.NoCredentialsError as e:
                logger.error(
                    " See https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html "
                    " for method of using environment variable to point to aws credentials, and the "
                    " order in which boto will search for said credentials. "
                )
                logger.error(
                    "Boto3 searches via the order below.  If on FAIR Cluster, method 4 may be most convenient."
                    ""
                    "The order in which Boto3 searches for credentials is:"
                    "1) [UNUSED] Passing credentials as parameters in the boto.client() method"
                    "2) [UNUSED] Passing credentials as parameters when creating a Session object"
                    "3) Environment variables"
                    "       AWS_ACCESS_KEY_ID - The access key for your AWS account."
                    "       AWS_SECRET_ACCESS_KEY - The secret key for your AWS account."
                    "       AWS_SESSION_TOKEN - The session key for your AWS account."
                    "           This is only needed when you are using temporary credentials. "
                    "4) Shared credential file (~/.aws/credentials)"
                    "       default: ~/.aws/credentials"
                    "       changed via: AWS_SHARED_CREDENTIALS_FILE"
                    "       *for FAIR cluster usage: `export AWS_SHARED_CREDENTIALS_FILE=~/.fairusers_aws/credentials`"
                    "5) AWS config file (~/.aws/config)"
                    "       default: ~/.aws/config"
                    "       changed via: AWS_CONFIG_FILE"
                    "6) Assume Role provider"
                    "7) Boto2 config file (/etc/boto.cfg and ~/.boto)"
                    "8) Instance metadata service on an Amazon EC2 instance that has an IAM role configured."
                )
                raise OSError(
                    f"Error in making s3 client for bucket {bucket}"
                    f"{type(e).__name__}: {e}"
                ) from e

        return self.client

    # pyre-fixme[3]: Return type must be annotated.
    def _local_cache_path(
        self,
        path: str,
    ):
        """
        Helper that returns a local cache path for a given uri.
        Args:
            path (str): A URI supported by this PathHandler.
        Returns:
            local_cache_path (str): a file path which exists on the local file system,
            in a cache directory.
        """
        bucket, file_path = self._parse_uri(path)
        return os.path.join(
            get_cache_dir(self.cache_dir), self.CACHE_SUBDIR_NAME, file_path
        )

    # pyre-fixme[14]: `_get_local_path` overrides method defined in `PathHandler`
    #  inconsistently.
    def _get_local_path(self, path: str, **kwargs: Any) -> str:
        """
        Get a filepath which is compatible with native Python I/O such as `open`
        and `os.path`.
        If URI points to a remote resource, this function may download and cache
        the resource to local disk. In this case, the cache stays on filesystem
        (under `file_io.get_cache_dir()`) and will be used by a different run.
        Therefore this function is meant to be used with read-only resources.
        Args:
            path (str): A URI supported by this PathHandler
        Returns:
            local_path (str): a file path which exists on the local file system
        """
        logger = logging.getLogger(__name__)
        self._check_kwargs(kwargs)

        # Cheap check first.
        if path.endswith("/"):
            raise NotImplementedError(
                "S3PathHandler does not currently support downloading directories"
            )
        assert self._isfile(path)

        local_path = self._local_cache_path(path)
        with file_lock(local_path):
            if os.path.exists(local_path):
                # If local object's last modified time is *after* remote object's last modified
                # time, do not use the cache.  Instead, redownload.
                response = self._head_object(path)
                if response is not None:
                    remote_dt = response["LastModified"]
                    local_dt = dt.datetime.fromtimestamp(
                        os.path.getmtime(local_path)
                    ).astimezone()
                    # NOTE: may consider still avoid cache if times are close, to avoid a race condition.
                    # Currently, a lengthy download of a very recent but stale file would have a late
                    # local last modified timestamp, and would be improperly used.
                    # Better fix: set last modified time via the remote object's last modified time,
                    # in download_file().
                    # pyre-fixme[58]: `>` is not supported for operand types
                    #  `datetime` and `timedelta`.
                    if (local_dt - remote_dt) > dt.timedelta(minutes=0):
                        logger.info(
                            "URL {} was already cached in {}".format(path, local_path)
                        )
                        return local_path

            logger.info("Caching {} ...".format(path))
            tmp = local_path + ".tmp"
            # clean-up tmp if found, because if tmp exists, it must be a dirty
            # result of a previously process that didn't cleanup itself.
            if os.path.isfile(tmp):
                os.unlink(tmp)

            bucket, s3_path = self._parse_uri(path)
            client = self._get_client(bucket)
            try:
                response = client.download_file(
                    bucket, s3_path, tmp, Config=self.transfer_config
                )

                # First download to tmp, then move it, because move is
                # (almost?) atomic when src and dst are in the same file
                # system. This will avoid partial cache state if the
                # process is killed.
                shutil.move(tmp, local_path)
            finally:
                try:
                    os.unlink(tmp)
                except Exception:
                    pass

            logger.info("URL {} cached in {}".format(path, local_path))
            return local_path

    # pyre-fixme[15]: `_copy_from_local` overrides method defined in `PathHandler`
    #  inconsistently.
    def _copy_from_local(
        self, local_path: str, dst_path: str, overwrite: bool = False, **kwargs: Any
    ) -> bool:
        """
        Copies a local file to the specified URI.
        If the URI is another local path, this should be functionally identical
        to copy.
        Args:
            local_path (str): a file path which exists on the local file system
            dst_path (str): A URI supported by this PathHandler
            overwrite (bool): Bool flag for forcing overwrite of existing URI
        Returns:
            status (bool): True on success
        """
        self._check_kwargs(kwargs)

        # Just checking this to avoid expensive API calls in self._isdir().
        if local_path.endswith("/") or dst_path.endswith("/"):
            raise NotImplementedError(
                "S3PathHandler does not currently support uploading directories"
            )

        if not overwrite and self._exists(dst_path):
            logger = logging.getLogger(__name__)
            logger.error("Error: Destination path {} already exists.".format(dst_path))
            return False

        bucket, s3_path = self._parse_uri(dst_path)
        client = self._get_client(bucket)
        try:
            client.upload_file(local_path, bucket, s3_path, Config=self.transfer_config)
            return True
        except botocore.exceptions.ClientError as e:
            logger = logging.getLogger(__name__)
            logger.error("Error in file upload - {}".format(str(e)))
            return False

    # pyre-fixme[3]: Return type must be annotated.
    def _decorate_buf_with_s3_methods(
        self,
        buffer: Union[IO[str], IO[bytes]],
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        client: Any,
        bucket: str,
        s3_path: str,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        transfer_config: Any,
    ):
        # Save old close method.
        # pyre-fixme[16]: `IO` has no attribute `_close`.
        buffer._close = buffer.close

        # Add in our new close method.
        fn = partial(
            s3_close_and_upload,
            client=client,
            bucket=bucket,
            s3_path=s3_path,
            transfer_config=transfer_config,
        )
        # pyre-fixme[8]: Attribute has type
        #  `BoundMethod[typing.Callable(IO.close)[[Named(self, IO[bytes])], None],
        #  IO[bytes]]`; used as `MethodType`.
        # pyre-fixme[8]: Attribute has type
        #  `BoundMethod[typing.Callable(IO.close)[[Named(self, IO[str])], None],
        #  IO[str]]`; used as `MethodType`.
        # pyre-fixme[9]: close has type
        #  `Union[BoundMethod[typing.Callable(IO.close)[[Named(self, IO[bytes])],
        #  None], IO[bytes]], BoundMethod[typing.Callable(IO.close)[[Named(self,
        #  IO[str])], None], IO[str]]]`; used as `MethodType`.
        buffer.close = types.MethodType(fn, buffer)

    def _open(
        self,
        path: str,
        mode: str = "r",
        buffering: int = -1,
        # The following three arguments are unused,
        # But are included to avoid triggering WARNING
        # messages from _check_kargs.
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
        read_chunk_size: Optional[int] = None,
        **kwargs: Any,
    ) -> Union[IO[str], IO[bytes]]:
        """
        Open a stream to a URI, similar to the built-in `open`.
        Args:
            path (str): A URI supported by this PathHandler
            mode (str): Specifies the mode in which the file is opened. It defaults
                to 'r'.
            buffering (int): An optional integer used to set the buffering policy.
                Pass 0 to switch buffering off and an integer >= 1 to indicate the
                size in bytes of a fixed-size chunk buffer. When no buffering
                argument is given, the default buffering policy depends on the
                underlying I/O implementation.
        Returns:
            file: a file-like object.
        """
        self._check_kwargs(kwargs)

        bucket, s3_path = self._parse_uri(path)
        client = self._get_client(bucket)

        # AWS methods download_fileobj() and upload_fileobj()
        # both expect binary file-like objects.
        if "r" in mode:
            if read_chunk_size is None:
                # 1. Download into io.BytesIO.
                # (binary format is required by download_fileobj.)
                buffer = io.BytesIO()
                try:
                    # NOTE: Will download entire file!  Further optimization to
                    # only read a portion of the file could be implemented here.
                    # NOTE: We download into an in-memory buffer.  If downloading to
                    # filesystem is desirable, use _get_local_path().
                    client.download_fileobj(
                        bucket, s3_path, buffer, Config=self.transfer_config
                    )
                except botocore.exceptions.ClientError as e:
                    raise OSError(
                        f"Error in making s3 client for bucekt {bucket}"
                        f"{type(e).__name__}: {e}"
                    ) from e

                # 2. Set file-pointer to beginning of file.
                buffer.seek(0)
            else:
                # pyre-fixme[6]: For 3rd param expected `int` but got `str`.
                buffer = S3ChunkReadIO(client, bucket, s3_path, read_chunk_size)
            # pyre-fixme[16]: `S3PathHandler` has no attribute `length`.
            self.length = client.get_object(Bucket=bucket, Key=s3_path)["ContentLength"]

            # 3. Use convenient wrapper to make object look like StringIO,
            # if user wants non-binary.
            encoding = None

            if "b" not in mode:
                encoding = "utf-8"
                return io.TextIOWrapper(
                    # pyre-fixme[6]: For 1st param expected `IO[bytes]` but got
                    #  `Union[S3ChunkReadIO, BytesIO]`.
                    buffer,
                    write_through=True,
                    encoding=encoding,
                    errors=None,
                    newline=None,
                    line_buffering=False,
                )
            else:
                # pyre-fixme[7]: Expected `Union[IO[bytes], IO[str]]` but got
                #  `Union[S3ChunkReadIO, BytesIO]`.
                return buffer

        elif "w" in mode:
            # 1. For writing, we give the user io.BytesIO or io.StringIO.
            if "b" in mode:
                buffer = io.BytesIO()
            else:
                buffer = io.StringIO()

            # 2. Decorate buffer so that we upload when it's closed by user.
            #       If StringIO, decorator does a simple+expensive conversion
            #       to bytesIO before uploading.
            #       (because upload_fileobj requires binary)
            self._decorate_buf_with_s3_methods(
                buffer, client, bucket, s3_path, self.transfer_config
            )

            return buffer

        else:
            raise OSError(f"Unsupported open mode {mode}")

    def _copy(
        self, src_path: str, dst_path: str, overwrite: bool = False, **kwargs: Any
    ) -> bool:
        """
        Copies a source path to a destination path.
        Args:
            src_path (str): A URI supported by this PathHandler
            dst_path (str): A URI supported by this PathHandler
            overwrite (bool): Bool flag for forcing overwrite of existing file
        Returns:
            status (bool): True on success
        """
        self._check_kwargs(kwargs)

        if not overwrite and self._exists(dst_path):
            logger = logging.getLogger(__name__)
            logger.error("Error: Destination path {} already exists.".format(dst_path))
            return False

        src_bucket, src_s3_path = self._parse_uri(src_path)
        dst_bucket, dst_s3_path = self._parse_uri(dst_path)
        assert src_bucket == dst_bucket, "For now, can only _copy() within a bucket."
        client = self._get_client(src_bucket)

        try:
            client.copy(
                {
                    "Bucket": src_bucket,
                    "Key": src_s3_path,
                },
                dst_bucket,
                dst_s3_path,
                Config=self.transfer_config,
            )
            return True
        except botocore.exceptions.ClientError as e:
            logger = logging.getLogger(__name__)
            logger.error("Error in file copy - {}".format(str(e)))
            return False

    # pyre-fixme[24]: Generic type `dict` expects 2 type parameters, use
    #  `typing.Dict` to avoid runtime subscripting errors.
    def _head_object(self, path: str) -> Optional[Dict]:
        bucket, s3_path = self._parse_uri(path)
        client = self._get_client(bucket)

        try:
            # Raises exception if not exists, else it exists.
            response = client.head_object(Bucket=bucket, Key=s3_path)
            return response
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Message"] == "Bad Request":
                raise OSError(
                    f"Error in checking s3 path {path} - " f"{type(e).__name__}: {e}"
                ) from e
            return None

    def _exists(self, path: str, **kwargs: Any) -> bool:
        """
        Checks if there is a resource at the given URI.
        Args:
            path (str): A URI supported by this PathHandler
        Returns:
            bool: true if the path exists
        """
        self._check_kwargs(kwargs)

        return self._head_object(path) is not None

    def _isfile(self, path: str, **kwargs: Any) -> bool:
        """
        Checks if the resource at the given URI is a file.
        Args:
            path (str): A URI supported by this PathHandler
        Returns:
            bool: true if the path is a file
        """
        self._check_kwargs(kwargs)

        # NOTE: this incurs an API call.
        return not path.endswith("/") and self._exists(path, **kwargs)

    def _isdir(self, path: str, **kwargs: Any) -> bool:
        """
        Checks if the resource at the given URI is a directory.
        Args:
            path (str): A URI supported by this PathHandler
        Returns:
            bool: true if the path is a directory
        """
        self._check_kwargs(kwargs)

        # NOTE: this incurs an API call.
        return path.endswith("/") and self._exists(path, **kwargs)

    def _ls(self, path: str, **kwargs: Any) -> List[str]:
        """
        List the contents of the directory at the provided URI.
        Args:
            path (str): A URI supported by this PathHandler
        Returns:
            List[str]: list of contents in given path
        """
        self._check_kwargs(kwargs)

        bucket, s3_path = self._parse_uri(path)
        client = self._get_client(bucket)

        try:
            # Pagination needed if >1000 entries.
            paginator = client.get_paginator("list_objects_v2")
            pages = paginator.paginate(
                Bucket=bucket,
                Prefix=s3_path,
                Delimiter="/",
            )
            obj_results = [
                obj["Key"] for page in pages for obj in page.get("Contents", [])
            ]
            dir_results = [
                obj["Prefix"]
                for page in pages
                for obj in page.get("CommonPrefixes", [])
            ]
            return obj_results + dir_results
        except botocore.exceptions.ClientError as e:
            raise OSError(
                f"Error in ls path {path} - " f"{type(e).__name__}: {e}"
            ) from e

    def _mkdirs(self, path: str, **kwargs: Any) -> None:
        """
        Recursive directory creation function. Like mkdir(), but makes all
        intermediate-level directories needed to contain the leaf directory.
        Similar to the native `os.makedirs`.
        Args:
            path (str): A URI supported by this PathHandler
        """
        self._check_kwargs(kwargs)

        assert path.endswith("/"), path

        bucket, s3_path = self._parse_uri(path)
        client = self._get_client(bucket)

        try:
            client.put_object(Bucket=bucket, Key=s3_path)
        except botocore.exceptions.ClientError as e:
            raise OSError(
                f"Error in mkdirs path {path} - " f"{type(e).__name__}: {e}"
            ) from e

    def _rm(self, path: str, **kwargs: Any) -> None:
        """
        Remove the file (not directory) at the provided URI.
        Args:
            path (str): A URI supported by this PathHandler
        """
        self._check_kwargs(kwargs)

        bucket, s3_path = self._parse_uri(path)
        client = self._get_client(bucket)

        try:
            client.delete_object(Bucket=bucket, Key=s3_path)
        except botocore.exceptions.ClientError as e:
            raise OSError(
                f"Error in rm path {path} - " f"{type(e).__name__}: {e}"
            ) from e


class S3ChunkReadIO(io.BufferedIOBase):
    # pyre-fixme[4]: Attribute must be annotated.
    DEFAULT_CHUNK_SIZE = 50 * 1024 * 1024  # 50MB

    # pyre-fixme[3]: Return type must be annotated.
    def __init__(
        self,
        # pyre-fixme[2]: Parameter must be annotated.
        client,
        bucket: str,
        key: int,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        timeout: Optional[timedelta] = None,
    ):
        # pyre-fixme[4]: Attribute must be annotated.
        self.client = client
        self.bucket = bucket
        self.key = key
        # pyre-fixme[4]: Attribute must be annotated.
        self.timeout = timeout.total_seconds() if timeout is not None else None
        self.chunk_size = chunk_size
        self.offset = 0
        self.buffered_window = range(0, 0)
        self.buffer = io.BytesIO()
        # pyre-fixme[4]: Attribute must be annotated.
        self.length = client.get_object(Bucket=bucket, Key=key)["ContentLength"]

    @property
    def name(self) -> str:
        # pyre-fixme[16]: `S3ChunkReadIO` has no attribute `path`.
        return self.path

    def seekable(self) -> bool:
        """
        Return a bool indicating whether object supports random access.

        If False, seek(), tell() and truncate() will raise OSError.
        This method may need to do a test seek().
        """
        return True

    def readable(self) -> bool:
        """
        Return a bool indicating whether object was opened for reading.

        If False, read() will raise OSError.
        """
        return True

    def writable(self) -> bool:
        """
        Return a bool indicating whether object was opened for writing.

        If False, write() and truncate() will raise OSError.
        """
        return False

    def fileno(self) -> int:
        raise AttributeError()

    def seek(self, offset: int, whence: int = 0) -> int:
        """
        Change stream position.

        Change the stream position to byte offset offset. Argument offset is
        interpreted relative to the position indicated by whence.  Values
        for whence are ints:

        * 0 -- start of stream (the default); offset should be zero or positive
        * 1 -- current stream position; offset may be negative
        * 2 -- end of stream; offset is usually negative
        Some operating systems / file systems could provide additional values.

        Return an int indicating the new absolute position.
        """
        if whence == 0:
            assert offset >= 0
            self.offset = offset
        elif whence == 1:
            assert offset + self.offset >= 0
            self.offset += offset
        elif whence == 2:
            self.offset = self.length + offset
        return self.offset

    def tell(self) -> int:
        """Return an int indicating the current stream position."""
        return self.offset

    def truncate(self, size: Optional[int] = None) -> int:
        """
        Truncate file to size bytes.

        Size defaults to the current IO position as reported by tell().  Return

        the new size.
        """
        raise OSError("can't truncate readonly stream")

    # pyre-fixme[14]: `write` overrides method defined in `BufferedIOBase`
    #  inconsistently.
    # pyre-fixme[15]: `write` overrides method defined in `BufferedIOBase`
    #  inconsistently.
    def write(self, b: Union[bytes, bytearray]) -> Optional[int]:
        """
        Write bytes b to in-memory buffer, return number written.
        """
        raise OSError("can't write to readonly stream")

    def close(self) -> None:
        """
        noop
        """
        pass

    def read1(self, size: int = -1) -> bytes:
        return self.read(size)

    # pyre-fixme[14]: `read` overrides method defined in `BufferedIOBase`
    #  inconsistently.
    def read(self, size: int = -1) -> bytes:
        """
        Read and return up to size bytes. If the argument is omitted, None, or negative,
        data is read and returned until EOF is reached. An empty bytes object is
        returned if the stream is already at EOF.
        """

        if size is None or size < 0:
            size = self.length - self.offset

        size = min(size, self.length - self.offset)

        ret = bytearray()

        if self.offset in self.buffered_window:
            buffer_offset = self.offset - self.buffered_window.start
            ret += self.buffer.getbuffer()[
                buffer_offset : min(buffer_offset + size, len(self.buffered_window))
            ]

        # if we already get enough data, return
        if len(ret) == size:
            self.offset += len(ret)
            return bytes(ret)

        # if partial data is available in the buffer, get the remaining data from S3
        if size - len(ret) > self.chunk_size:
            self.offset += len(ret)
            # For s3, range x-x means 1 byte at offset x
            output = self._read_from_s3(
                range(self.offset, min(self.offset + size - len(ret) - 1, self.length))
            )
            self.offset += len(output)
            # pyre-fixme[7]: Expected `bytes` but got `bytearray`.
            return ret + output

        # otherwise download the next chunk from s3, update buffer and buffered window
        self._read_chunk_to_buffer(self.offset + len(ret))

        # append the remaining data from newly downloaded buffer and return
        ret += self.buffer.getbuffer()[0 : size - len(ret)]

        assert len(ret) == size
        self.offset += len(ret)
        return bytes(ret)

    def _read_from_s3(self, download_range: range) -> bytes:
        obj = self.client.get_object(
            Bucket=self.bucket,
            Key=self.key,
            Range=f"bytes={download_range.start}-{download_range.stop}",
        )
        streaming_body = obj["Body"]

        if self.timeout is not None:
            streaming_body.set_socket_timeout(self.timeout)

        ret = bytearray()
        for chunk in streaming_body.iter_chunks(chunk_size=self.chunk_size):
            ret += chunk
        streaming_body.close()
        # pyre-fixme[7]: Expected `bytes` but got `bytearray`.
        return ret

    def _read_chunk_to_buffer(self, start_offset: int) -> None:
        """
        download a chuck size of data start from start_offset into current buffer, then update
        self.buffered_window for booking which part of data is currently buffered
        """
        download_range = range(
            start_offset, min(start_offset + self.chunk_size, self.length)
        )

        ret = self._read_from_s3(download_range)

        self.buffer.seek(0)
        self.buffer.write(ret)
        self.buffered_window = download_range



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
    Utility for testing async writes with `NativePathHandler`.
    Usage:
        buck run @mode/opt //fair_infra/data/iopath/tests:async_writes_test
"""

import logging
import os
import tempfile
import time

from iopath.common.file_io import PathManager


# pyre-fixme[5]: Global expression must be annotated.
logger = logging.getLogger(__name__)


# pyre-fixme[3]: Return type must be annotated.
# pyre-fixme[2]: Parameter must be annotated.
def printx(str):
    logger.warning(f"[{time.strftime('%X')}] {str}")


class TestDriver:
    LEN = 100000000  # This many characters per append job
    NUM_JOBS = 10
    _pathmgr = PathManager()

    def test(self) -> None:
        with tempfile.TemporaryDirectory() as _tmpdir:
            URI = os.path.join(_tmpdir, "test.txt")

            start_time = time.time()
            printx(
                f"Start dispatching {self.NUM_JOBS} async write jobs "
                f"each with {self.LEN} characters"
            )

            FINAL_STR = ""
            with self._pathmgr.opena(URI, "a") as f:
                for i in range(self.NUM_JOBS):  # `i` goes from 0 to 9
                    FINAL_STR += f"{i}" * self.LEN
                    f.write(f"{i}" * self.LEN)

            mid_time = time.time()
            printx(
                f"Time taken to dispatch {self.NUM_JOBS} threads: {mid_time - start_time}"
            )
            printx("Calling `async_join()`")
            # We want this `async_join` call to take time. If it is instantaneous, then our
            # async write calls are not running asynchronously.
            assert self._pathmgr.async_join()
            printx(
                f"Time Python waited for `async_join()` call to finish: {time.time() - mid_time}"
            )

            assert self._pathmgr.async_close()

            with self._pathmgr.open(URI, "r") as f:
                assert f.read() == FINAL_STR

            printx("Async Writes Test finish.")
            printx(
                "Passing metric: "
                "If the `async_join()` call took more than a negligible time to complete, "
                "then Python waited for the threads to finish and the Async Writes "
                "Test SUCCEEDS. Otherwise FAILURE."
            )


if __name__ == "__main__":
    printx("Async Writes Test starting.")
    tst = TestDriver()
    tst.test()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import runpy

from setuptools import find_packages, setup


version = runpy.run_path("iopath/version.py")["__version__"]

setup(
    name="iopath",
    version=version,
    author="FAIR",
    license="MIT licensed, as found in the LICENSE file",
    url="https://github.com/facebookresearch/iopath",
    description="A library for providing I/O abstraction.",
    python_requires=">=3.6",
    install_requires=[
        "tqdm",
        "typing_extensions",
        "portalocker",
        "dataclasses; python_version<'3.7'",
    ],
    extras_require={
        "aws": [
            "boto3",
        ],
    },
    packages=find_packages(exclude=("tests",)),
)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import Any

from iopath.common.file_io import PathHandler, TabularIO


class TabularUriParser:
    def parse_uri(self, uri: str) -> None:
        pass


class TabularPathHandler(PathHandler):
    def _opent(
        self, path: str, mode: str = "r", buffering: int = 32, **kwargs: Any
    ) -> TabularIO:
        # pyre-fixme[7]: Expected `TabularIO` but got implicit return value of `None`.
        assert mode == "r"
